"""
Jobs
====

Run arbitrary scripts on NimbleBox.ai instances with simplicity of ``nbox``.

This immensely powerful tool will allow methods to be run in a truly serverless fashion.
In the background it will spin up a VM, then run a script and close the machine. This is
the second step towards YoCo and CasH, read more
`here <https://yashbonde.github.io/general-perceivers/remote.html>`_.
"""

import os
import re
import sys
import time
from functools import partial
from tabulate import tabulate
from requests.sessions import Session

from .utils import SpecSubway, Subway, TIMEOUT_CALLS
from ..utils import nbox_session, NBOX_HOME_DIR, join
from ..auth import secret

from logging import getLogger
logger = getLogger()


def get_status(url = "https://nimblebox.ai", session = nbox_session):
  r = session.get(f"{url}/api/instance/get_user_instances")
  r.raise_for_status()
  message = r.json()["msg"]
  if message != "success":
    raise Exception(message)

  money = r.json()["nbBucks"]
  data = [{k: x[k] for k in Instance.useful_keys} for x in r.json()["data"]]
  return money, data

def print_status(url = "https://nimblebox.ai"):
  money, data = get_status(url)
  logger.info(f"Total NimbleBox.ai credits left: {money}")
  data_table = [[x[k] for k in Instance.useful_keys] for x in data]
  for x in tabulate(data_table, headers=Instance.useful_keys).splitlines():
    logger.info(x)

def get_instance(url, id_or_name, session = nbox_session):
  if not isinstance(id_or_name, (int, str)):
    raise ValueError("Instance id must be an integer or a string")

  r = session.get(f"{url}/api/instance/get_user_instances")
  r.raise_for_status()
  resp = r.json()

  if len(resp["data"]) == 0:
    raise ValueError(f"No instance: '{id_or_name}' found, create manually from the dashboard or Instance.new(...)")

  key = "instance_id" if isinstance(id_or_name, int) else "name"
  instance = list(filter(lambda x: x[key] == id_or_name, resp["data"]))
  if len(instance) == 0:
    raise KeyError(id_or_name)
  instance = instance[0] # pick the first one

  return instance

def is_random_name(name):
  return re.match(r"[a-z]+-[a-z]+", name) is not None


class Instance():
  # each instance has a lot of data against it, we need to store only a few as attributes
  useful_keys = ["state", "used_size", "total_size", "public", "instance_id", "name"]

  def __init__(self, id_or_name, loc = None, cs_endpoint = "server"):
    super().__init__()

    self.url = f"https://{'' if not loc else loc+'.'}nimblebox.ai"
    self.cs_url = None
    self.cs_endpoint = cs_endpoint
    self.session = Session()
    self.session.headers.update({"Authorization": f"Bearer {secret.get('access_token')}"})
    self.web_server = Subway(f"{self.url}/api/instance", self.session)
    logger.info(f"WS: {self.web_server}")

    self.instance_id = None
    self.__opened = False
    self.running_scripts = []

    self.refresh(id_or_name)
    logger.info(f"Instance added: {self.name} ({self.instance_id})")

    if self.state == "RUNNING":
      self.start()

  def __eq__(self, __o: object):
    return self.instance_id == __o.instance_id

  def __repr__(self):
    return f"<Instance ({', '.join([f'{k}:{getattr(self, k)}' for k in self.useful_keys + ['cs_url']])})>"

  # normally all the methods that depend on a certain source (eg. webserver) are defined in ext.py
  # file that has all the extensions. However because create is a special method we are defining it
  # as a class method.
  @classmethod
  def create(cls, name, url = "https://nimblebox.ai"):
    r = nbox_session.post(
      f"{url}/api/instance/create_new_instance_v4",
      json = {"project_name": name, "project_template": "blank"}
    )
    r.raise_for_status() # if its not 200, it's an error
    return cls(name, url)

  def start(self, cpu_only = True, cpu_count = 2, gpu = "p100", gpu_count = 1):
    """``cpu_count`` should be one of [2, 4, 8]"""
    if self.__opened:
      logger.info(f"Instance {self.name} ({self.instance_id}) is already opened")
      return

    if not self.state == "RUNNING":
      logger.info(f"Starting instance {self.name} ({self.instance_id})")
      message = self.web_server.start_instance(
        "post",
        data = {
          "instance_id": self.instance_id,
          "hw":"cpu" if cpu_only else "gpu",
          "hw_config":{
            "cpu":f"n1-standard-{cpu_count}",
            "gpu":f"nvidia-tesla-{gpu}",
            "gpuCount": gpu_count,
          }
        }
      )["msg"]
      if not message == "success":
        raise ValueError(message)

      logger.info(f"Waiting for instance {self.name} ({self.instance_id}) to start")
      _i = 0
      while self.state != "RUNNING":
        time.sleep(5)
        self.refresh()
        _i += 1
        if _i > TIMEOUT_CALLS:
          raise TimeoutError("Instance did not start within timeout, please check dashboard")
      logger.info(f"Instance {self.name} ({self.instance_id}) started")
    else:
      logger.info(f"Instance {self.name} ({self.instance_id}) is already running")

    # now the instance is running, we can open it, opening will assign a bunch of cookies and
    # then get us the exact location of the instance
    logger.info(f"Opening instance {self.name} ({self.instance_id})")
    self.open_data = self.web_server.open_instance(
      "post", data = {"instance_id":self.instance_id}
    )
    instance_url = self.open_data["base_url"].lstrip("/").rstrip("/")
    self.cs_url = f"{self.url}/{instance_url}"
    if self.cs_endpoint:
      self.cs_url += f"/{self.cs_endpoint}"

    # TODO:@yashbonde remove this check, once v2 on prod
    if self.cs_url.endswith("server"):
      self.__opened = True
      return

    # create a speced-subway, this requires capturing the openAPI spec first
    r = self.session.get(f"{self.cs_url}/openapi.json"); r.raise_for_status()
    self.cs_spec = r.json()
    self.compute_server = SpecSubway.from_openapi(self.cs_spec, self.cs_url, self.session)
    logger.info(f"CS: {self.compute_server}")

    # now load all the functions from methods.py
    logger.info(f"Testing instance {self.name} ({self.instance_id})")
    out = self.compute_server.test()
    with open(join(NBOX_HOME_DIR, "methods.py"), "w") as f:
      f.write(out["data"])

    sys.path.append(NBOX_HOME_DIR)
    from methods import __all__ as m_all
    for m in m_all:
      trg_name = f"bios_{m}"
      exec(f"from methods import {m} as {trg_name}")
      fn = partial(
        locals()[trg_name],
        cs_sub = self.compute_server,
        ws_sub = self.web_server,
        logger = logger
      )
      setattr(self, m, fn)

    # see if any developmental methods are saved locally
    try:
      from proto_methods import __all__ as p_all
      for m in p_all:
        trg_name = f"bios_{m}"
        exec(f"from proto_methods import {m} as {trg_name}")
        fn = partial(
          locals()[trg_name],
          cs_sub = self.compute_server,
          ws_sub = self.web_server,
          logger = logger
        )
        setattr(self, m, fn)
    except:
      pass

    self.__opened = True

  def stop(self):
    if self.state == "STOPPED":
      logger.info(f"Instance {self.name} ({self.instance_id}) is already stopped")
      return

    logger.info(f"Stopping instance {self.name} ({self.instance_id})")
    message = self.web_server.stop_instance("post", data = {"instance_id":self.instance_id})["msg"]
    if not message == "success":
      raise ValueError(message)

    logger.info(f"Waiting for instance {self.name} ({self.instance_id}) to stop")
    _i = 0 # timeout call counter
    while self.state != "STOPPED":
      time.sleep(5)
      self.refresh()
      _i += 1
      if _i > TIMEOUT_CALLS:
        raise TimeoutError("Instance did not stop within timeout, please check dashboard")
    logger.info(f"Instance {self.name} ({self.instance_id}) stopped")

    self.__opened = False

  def delete(self, force = False):
    if self.__opened and not force:
      raise ValueError("Instance is still opened, please call .stop() first")
    logger.info(f"Deleting instance {self.name} ({self.instance_id})")
    message = self.web_server.delete_instance("post", data = {"instance_id":self.instance_id})["msg"]
    if not message == "success":
      raise ValueError(message)

  def __call__(self, x):
    """Caller is the most important UI/UX. The letter ``x`` in programming is reserved the most
    arbitrary thing, and this ``nbox.Instance`` is the gateway to a cloud instance. You can:
    1. run a script on the cloud
    2. run a local script on the cloud
    3. get the status of a script
    4. [TBD] run a special kind of functions known as
    `pure functions <https://en.wikipedia.org/wiki/Pure_function>`_

    Pure functions in programming are functions that are self sufficient in terms of execution,
    eg. all the packages are imported inside the function and there are no side effects in an
    execution (seeds included for probabilistic functions). Writing such functions in python with
    any IDE with checker is dead easy, however the performace guarantee is premium given high
    costs. Thus the logic to parsing these will have to be written as a seperate module.

    Returns:
      [type]: [description]
    """
    if not self.__opened:
      raise ValueError("Instance is not opened, please call .start() first")
    if not isinstance(x, str):
      raise ValueError("x must be a string")

    def _run_cloud(fpath):
      fpath = "/mnt/disks/user/project/" + fpath.split("nbx://")[1]
      meta = list(filter(
        lambda y: y["path"] == fpath, self.compute_server.myfiles.info(fpath)["data"]
      ))
      if not meta:
        raise ValueError(f"File {x} not found on instance {self.name} ({self.instance_id})")
      return self.compute_server.rpc.start(fpath)["uid"]

    if is_random_name(x):
      logger.info(f"Getting status of Job '{x}' on instance {self.name} ({self.instance_id})")
      data = self.compute_server.rpc.status(x)
      if not data["msg"] == "success":
        raise ValueError(data["msg"])
      else:
        status = data["status"]
        logger.info(f"Script {x} on instance {self.name} ({self.instance_id}) is {status}")

      # if stopped then get the logs and run
      if status == "stopped":
        out = self.compute_server.rpc.logs(x)
        if out["err"]:
          logger.error("Execution errored out")
          return "error-done"
        else:
          return "done"
      
      return status
    elif os.path.isfile(x):
      self.mv(x, f"nbx://{x}")
      uid = _run_cloud(f"nbx://{x}")
      self.running_scripts.append(uid)
      return uid
    elif x.startswith("nbx://"):
      logger.info("Running file on cloud")
      uid = _run_cloud(x)
      self.running_scripts.append(uid)
      return uid
    else:
      raise ValueError(f"Unknown: {x}")

  def refresh(self, id_or_name = None):
    id_or_name = id_or_name or self.instance_id
    out = get_instance(self.url, id_or_name, session = self.session)
    for k,v in out.items():
      if k in self.useful_keys:
        setattr(self, k, v)
    self.data = out

    if self.state == "RUNNING":
      self.start()



# this file has bunch of functions that are used everywhere

import os
import io
import hashlib
import requests
import tempfile
import randomname
from uuid import uuid4

import logging


# common/
nbox_session = requests.Session()
# /common

# logging/
logger = logging.getLogger()
# /logging


# lazy_loading/

def isthere(*packages, soft = True):
    def wrapper(fn):
        def _fn(*args, **kwargs):
            # since we are lazy evaluating this thing, we are checking when the function
            # is actually called. This allows checks not to happen during __init__.
            for package in packages:
                try:
                    __import__(package)
                except ImportError:
                    if not soft:
                        raise Exception(f"{package} is not installed")
                    # raise a warning, let the modulenotfound exception bubble up
                    logger.warn(
                        f"{package} is not installed, but is required by {fn.__module__}, some functionality may not work"
                    )
            return fn(*args, **kwargs)
        return _fn
    return wrapper

def _isthere(*packages):
    for package in packages:
        try:
            __import__(package)
        except Exception:
            return False
    return True

# /lazy_loading

# file path/reading

def fetch(url):
    # efficient loading of URLs
    fp = os.path.join(tempfile.gettempdir(), hash_(url))
    if os.path.isfile(fp) and os.stat(fp).st_size > 0:
        with open(fp, "rb") as f:
            dat = f.read()
    else:
        dat = requests.get(url).content
        with open(fp + ".tmp", "wb") as f:
            f.write(dat)
        os.rename(fp + ".tmp", fp)
    return dat


def folder(x):
    # get the folder of this file path
    return os.path.split(os.path.abspath(x))[0]

def join(x, *args):
    return os.path.join(x, *args)

NBOX_HOME_DIR = join(os.path.expanduser("~"), ".nbx")

# /path

# misc/

def get_random_name(uuid = False):
    if uuid:
        return str(uuid4())
    return randomname.generate()

def hash_(item, fn="md5"):
    return getattr(hashlib, fn)(str(item).encode("utf-8")).hexdigest()

# /misc

# model/

def get_image(file_path_or_url):
    if os.path.exists(file_path_or_url):
        return Image.open(file_path_or_url)
    else:
        return Image.open(io.BytesIO(fetch(file_path_or_url)))

def convert_to_list(x):
    # recursively convert tensors -> list
    import torch
    if isinstance(x, list):
        return x
    if isinstance(x, dict):
        return {k: convert_to_list(v) for k, v in x.items()}
    elif isinstance(x, (torch.Tensor, np.ndarray)):
        x = np.nan_to_num(x, -1.42069)
        return x.tolist()
    else:
        raise Exception("Unknown type: {}".format(type(x)))

# /model

# pool/
from concurrent.futures import ThreadPoolExecutor, as_completed
POOL_SUPPORTED_MODES = ["thread"]

class Pool:
    def __init__(self, mode = "thread", max_workers = 2, _name: str = get_random_name(True)):
        """Threading is hard, your brain is not wired to handle parallelism. You are a blocking
        python program. So a blocking function for you.

        Args:
            mode (str, optional): There can be multiple pooling strategies across cores, threads,
                k8s, nbx-instances etc.
            max_workers (int, optional): Numbers of workers to use
            _name (str, optional): Name of the pool, used for logging

        Usage:
            
            def sleep_and_return(x):
                from time import sleep
                sleep(x)
                return x
            
            pool = Pool()
            results = pool(sleep_and_return, (2,),(6,),(4,),(5,)) # inputs must be a tuple
        """
        if mode not in POOL_SUPPORTED_MODES:
            raise Exception(f"Only {', '.join(POOL_SUPPORTED_MODES)} mode(s) are supported")

        self.mode = mode
        self.pool = None
        logger.info(f"Starting ThreadPool ({_name}) with {max_workers} workers")
        self.executor = ThreadPoolExecutor(
            max_workers=max_workers,
            thread_name_prefix=_name
        )
        self.item_id = -1 # because +1 later
        self.futures = {}

    def __call__(self, fn, *args):
        """Run any function ``fn`` in parallel, where each argument is a list of arguments to
        pass to ``fn``. Result is returned in the same order as the input.
            
            thread(fn, a) for a in args -> list of results
        """
        assert callable(fn)
        assert isinstance(args[0], (tuple, list))

        futures = {}
        for i, x in enumerate(args):
            futures[self.executor.submit(fn, *x)] = i # insertion index

        self.item_id += len(futures)
        results = {}
        for future in as_completed(futures):
            try:
                result = future.result()
                results[futures[future]] = result # update that index
            except Exception as e:
                logger.error(f"{self.mode} error: {e}")
                raise e

        return [results[x] for x in range(len(results))]

# /pool


# --- classes

# this needs to be redone
# # Console is a rich console wrapper for beautifying statuses
# class Console:
#     T = SimpleNamespace(
#         clk="deep_sky_blue1",  # timer
#         st="bold dark_cyan",  # status + print
#         fail="bold red",  # fail
#         inp="bold yellow",  # in-progress
#         nbx="bold bright_black",  # text with NBX at top and bottom
#         rule="dark_cyan",  # ruler at top and bottom
#         spinner="weather",  # status theme
#     )
# 
#     def __init__(self):
#         self.c = richConsole()
#         self._in_status = False
#         self.__reset()
# 
#     def rule(self, title: str):
#         self.c.rule(f"[{self.T.nbx}]{title}[/{self.T.nbx}]", style=self.T.rule)
# 
#     def __reset(self):
#         self.st = time()
# 
#     def __call__(self, x, *y):
#         cont = " ".join([str(x)] + [str(_y) for _y in y])
#         if not self._in_status:
#             self._log(cont)
#         else:
#             self._update(cont)
# 
#     def sleep(self, t: int):
#         for i in range(t):
#             self(f"Sleeping for {t-i}s ...")
#             _sleep(1)
# 
#     def _log(self, x, *y):
#         cont = " ".join([str(x)] + [str(_y) for _y in y])
#         t = str(timedelta(seconds=int(time() - self.st)))[2:]
#         self.c.print(f"[[{self.T.clk}]{t}[/{self.T.clk}]] {cont}")
# 
#     def start(self, x="", *y):
#         self.__reset()
#         cont = " ".join([str(x)] + [str(_y) for _y in y])
#         self.status = self.c.status(f"[{self.T.st}]{cont}[/{self.T.st}]", spinner=self.T.spinner)
#         self.status.start()
#         self._in_status = True
# 
#     def _update(self, x, *y):
#         t = str(timedelta(seconds=int(time() - self.st)))[2:]
#         cont = " ".join([str(x)] + [str(_y) for _y in y])
#         self.status.update(f"[[{self.T.clk}]{t}[/{self.T.clk}]] [{self.T.st}]{cont}[/{self.T.st}]")
# 
#     def stop(self, x):
#         self.status.stop()
#         del self.status
#         self._log(x)
#         self._in_status = False



# this file has the utilities and functions required for processing pytorch items
# such as conversion to ONNX, getting the metadata and so on.

import torch


def export_to_onnx(
    model,
    args,
    export_model_path,
    input_names,
    dynamic_axes,
    output_names,
    export_params=True,
    verbose=False,
    opset_version=12,
    do_constant_folding=True,
    use_external_data_format=False,
    **kwargs
):
    torch.onnx.export(
        model,
        args=args,
        f=export_model_path,
        input_names=input_names,
        verbose=verbose,
        output_names=output_names,
        use_external_data_format=use_external_data_format,  # RuntimeError: Exporting model exceed maximum protobuf size of 2GB
        export_params=export_params,  # store the trained parameter weights inside the model file
        opset_version=opset_version,  # the ONNX version to export the model to
        do_constant_folding=do_constant_folding,  # whether to execute constant folding for optimization
        dynamic_axes=dynamic_axes,
    )


def export_to_torchscript(model, args, export_model_path, **kwargs):
    traced_model = torch.jit.trace(model, args, check_tolerance=0.0001)
    torch.jit.save(traced_model, export_model_path)


__all__ = ["export_to_onnx", "export_to_torchscript"]



import joblib
from skl2onnx import convert_sklearn
import skl2onnx.common.data_types as dt

__NP_DTYPE_TO_SKL_DTYPE = {
    "bool_": dt.BooleanTensorType,
    "double": dt.DoubleTensorType,
    "int16": dt.Int16TensorType,
    "complex128": dt.Complex128TensorType,
    "int32": dt.Int32TensorType,
    "uint16": dt.UInt16TensorType,
    "complex64": dt.Complex64TensorType,
    "float16": dt.Float16TensorType,
    "int64": dt.Int64TensorType,
    "str_": dt.StringTensorType,
    "uint32": dt.UInt32TensorType,
    "float64": dt.FloatTensorType,
    "uint64": dt.UInt64TensorType,
    "int8": dt.Int8TensorType,
    "uint8": dt.UInt8TensorType,
}


def export_to_onnx(model, args, input_names, input_shapes, export_model_path, opset_version=None, **kwargs):
    # the args need to be converted to proper formatted data types
    initial_types = []
    for name, shape, tensor in zip(input_names, input_shapes, args):
        shape = list(shape)
        shape[0] = None  # batching requires the first dimension to be None
        initial_types.append((name, __NP_DTYPE_TO_SKL_DTYPE[str(tensor.dtype)](shape)))

    onx = convert_sklearn(model, initial_types=initial_types, target_opset=opset_version)

    with open(export_model_path, "wb") as f:
        f.write(onx.SerializeToString())


def export_to_pkl(model, export_model_path, **kwargs):
    # sklearn models are pure python methods (though underlying contains bindings to C++)
    # and so we can use joblib for this
    # we use the joblib instead of pickle
    with open(export_model_path, "wb") as f:
        joblib.dump(model, f)


__all__ = ["export_to_onnx", "export_to_pkl"]



from logging import getLogger
logger = getLogger()

TIMEOUT_CALLS = 60


class Subway():
    """Simple code that allows extending things by .attr.ing them"""
    def __init__(self, _url, _session):
        self._url = _url.rstrip('/')
        self._session = _session

    def __repr__(self):
        return f"<Subway ({self._url})>"

    def __getattr__(self, attr):
        return Subway(f"{self._url}/{attr}", self._session)

    def __call__(self, method = "get", trailing = "", data = None, _verbose = False):
        fn = getattr(self._session, method)
        url = f"{self._url}{trailing}"
        if _verbose:
            print(f"Calling {url}")
        r = fn(url, json = data)
        if _verbose:
            print(r.content.decode())
        r.raise_for_status() # good when server is good
        return r.json()


class SpecSubway():
    """Subway is simple when starting, but what if we had a spec that could perform sanity checks
    and parse data when __call__ ed. In this version it can read an OpenAPI spec and perform lazy
    checking"""
    def __init__(self, _url, _session, _spec, __name = None):
        self._url = _url.rstrip('/')
        self._session = _session
        self._spec = _spec
        self._name = __name
        self._caller = (
            (len(_spec) == 3 and set(_spec) == set(["method", "meta", "src"])) or
            (len(_spec) == 4 and set(_spec) == set(["method", "meta", "src", "response_kwargs_dict"])) or
            "/" in self._spec
        )

    @classmethod
    def from_openapi(cls, openapi, _url, _session):
        paths = openapi["paths"]
        spec = openapi["components"]["schemas"]
        
        tree = {}
        for p in tuple(paths.keys()):
            t = tree
            for part in p.split('/')[1:]:
                part = "/" if part == "" else part
                t = t.setdefault(part, {})

        def _dfs(tree, trail = []):
            for t in tree:
                if tree[t] == {}:
                    src = "/" + "/".join(trail)
                    if t!= "/":
                        src = src + "/" if src != "/" else src
                        src = src + t
                    
                    try:
                        data = paths[src]
                    except:
                        src = src + "/"
                        data = paths[src]
                    method = tuple(data.keys())[0]
                    body = data[method]
                    dict_ = {"method": method, "meta": None, "src": src}
                    if "requestBody" in body:
                        schema_ref = body["requestBody"]["content"]["application/json"]["schema"]["$ref"].split("/")[-1]
                        _req_body = spec[schema_ref]
                        kwargs_dict = list(_req_body["properties"])
                        dict_["meta"] = {
                            "kwargs_dict": kwargs_dict,
                            "required": _req_body.get("required", None)
                        }
                    if "responses" in body:
                        schema = body["responses"]["200"]["content"]["application/json"]["schema"]
                        if "$ref" in schema:
                            schema_ref = schema["$ref"].split("/")[-1]
                            _req_body = spec[schema_ref]
                            kwargs_dict = list(_req_body["properties"])
                            if dict_["meta"] != None:
                                dict_["meta"].update({"response_kwargs_dict": kwargs_dict})
                            else:
                                dict_["meta"] = {"response_kwargs_dict": kwargs_dict}
                    tree[t] = dict_
                else:
                    _dfs(tree[t], trail + [t])

        _dfs(tree)
        
        return cls(_url, _session, tree)

    def __repr__(self):
        return f"<SpecSubway ({self._url})>"

    def __getattr__(self, attr):
        # https://stackoverflow.com/questions/3278077/difference-between-getattr-vs-getattribute
        if self._caller and len(self._spec) == 1:
            raise AttributeError(f"'.{self._name}' does not have children")
        if attr not in self._spec:
            raise AttributeError(f"'.{attr}' is not a valid function")
        return SpecSubway(f"{self._url}/{attr}", self._session, self._spec[attr], attr)
    
    def __call__(self, *args, _verbose = False, _parse = False, **kwargs):
        if not self._caller:
            raise AttributeError(f"'.{self._name}' is not an endpoint")
        spec = self._spec
        if self._caller and "/" in self._spec:
            spec = self._spec["/"]
        
        data = None
        if spec["meta"] == None:
            assert len(args) == len(kwargs) == 0, "This method does not accept any arguments"
        else:
            spec_meta = spec["meta"]
            if "kwargs_dict" not in spec_meta:
                assert len(args) == len(kwargs) == 0, "This method does not accept any arguments"
            else:
                kwargs_dict = spec["meta"]["kwargs_dict"]
                required = spec["meta"]["required"]
                data = {}
                for i in range(len(args)):
                    if required != None:
                        data[required[i]] = args[i]
                    else:
                        data[kwargs_dict[i]] = args[i]
                for key in kwargs:
                    if key not in kwargs_dict:
                        raise AttributeError(f"{key} is not a valid argument")
                    data[key] = kwargs[key]
                if required != None:
                    for key in required:
                        if key not in data:
                            raise AttributeError(f"{key} is a required argument")

        fn = getattr(self._session, spec["method"])
        url = f"{self._url}"
        if self._caller and "/" in self._spec:
            url += "/"
        if _verbose:
            print(f"{spec['method'].upper()} {url}")
            print("-->>", data)
        r = fn(url, json = data)
        if not r.status_code == 200:
            raise ValueError(r.content.decode())
        
        out = r.json()
        if _parse and self._spec["meta"] != None and "response_kwargs_dict" in self._spec["meta"]:
            out = [out[k] for k in self._spec["meta"]["response_kwargs_dict"]]
            if len(out) == 1:
                return out[0]
        return out




import os

from . import utils

import logging
logger = logging.getLogger()


# ----- parsers
# These objects are mux, they consume and streamline the output
# Don't know what mux are? Study electronics.


class BaseParser:
    """This is the base parser class which has the required methods to process
    the three kind of structures:

    #. None
    #. list: ``process_list()``
    #. dict: ``process_dict()``

    and a function dedicated to processing the primitives

    #. ``process_primitive()``

    These functions are called from the ``__call__`` method.
    """

    def __init__(self):
        pass

    def process_primitive(self):
        raise NotImplementedError

    def process_dict(self):
        raise NotImplementedError

    def process_list(self):
        raise NotImplementedError

    def __call__(self, input_object):
        if isinstance(input_object, list):
            logger.info(f" - {self.__class__.__name__} - (A1)")
            out = self.process_list(input_object)
        elif isinstance(input_object, dict):
            logger.info(f" - {self.__class__.__name__} - (A2)")
            out = self.process_dict(input_object)
        else:
            logger.info(f" - {self.__class__.__name__} - (A3)")
            out = self.process_primitive(input_object)

        # apply self.post_proc_fn if it exists
        if self.post_proc_fn:
            # either you are going to get a dict or you are going to get a np.ndarray
            if isinstance(out, dict):
                if isinstance(next(iter(out.values())), dict):
                    # dict of dicts kinda thingy (say when two inputs are given)
                    out = {k: {_k: self.post_proc_fn(_v) for _k, _v in v.items()} for k, v in out.items()}
                else:
                    out = {k: self.post_proc_fn(v) for k, v in out.items()}
            else:
                out = self.post_proc_fn(out)
        return out


# ----- parsers for each category


class ImageParser(BaseParser):
    def __init__(self, post_proc_fn=None, cloud_infer=False, **kwargs):
        """single unified Image parser that consumes different types of data and returns a processed numpy array

        Args:
            post_proc_fn (callable, optional): post processing function, this takes in the torch tensor and performs
                operation
            cloud_infer (bool, optional): whether the input is from cloud inference
            kwargs (dict, optional): keyword arguments to store here
        """
        super().__init__()
        self.post_proc_fn = post_proc_fn
        self.cloud_infer = cloud_infer
        for k, v in kwargs.items():
            setattr(self, k, v)

    # common operations
    # if not cloud_infer and input is int then rescale it 0, 1
    def rescale(self, x):
        if not self.cloud_infer and "int" in str(x.dtype):
            x = x / 255
        return x

    def rearrange(self, x):
        if len(x.shape) == 3 and x.shape[0] != 3:
            return x.transpose(2, 0, 1)
        elif len(x.shape) == 4 and x.shape[1] != 3:
            return x.transpose(0, 3, 1, 2)
        return x

    def process_primitive(self, x, target_shape=None):
        """primitive can be string, array, Image"""
        if isinstance(x):
            logger.info(" - ImageParser - (C2) np.ndarray")
            # if shape == 3, unsqueeze it, numpy arrays cannot be reshaped
            out = x[None, ...] if len(x.shape) == 3 else x
        elif isinstance(x, Image.Image):
            logger.info(" - ImageParser - (C1) Image.Image object")
            img = x
        elif isinstance(x, str):
            if os.path.isfile(x):
                logger.info(" - ImageParser - (C3) string - file")
                img = Image.open(x)
            elif x.startswith("http"):
                logger.info(" - ImageParser - (C4) string - url")
                img = utils.get_image(x)
            else:
                try:
                    # probably base64
                    from io import BytesIO
                    import base64

                    logger.info(" - ImageParser - (C5) string - base64")
                    img = Image.open(BytesIO(base64.b64decode(x)))
                except:
                    raise Exception("Unable to parse string as Image")
        else:
            raise ValueError("Unknown primitive type: {}".format(type(x)))

        if not isinstance(x, np.ndarray):
            img = img.convert("RGB")

            # this checks when only the primitive is sent directly but there is just one template
            # so just use the only shape it has
            if target_shape is None and hasattr(self, "templates") and len(self.templates) == 1:
                target_shape = self.templates[next(iter(self.templates.keys()))]
                target_shape = target_shape[-2:][::-1]  # [h,w] -> [w,h]

            # if a certain target shape is given, then resize it to that shape
            if target_shape is not None:
                target_shape = target_shape
                img = img.resize(target_shape)

            out = self.process_primitive(np.array(img))

        # finally perform rearrange and rescale
        return self.rescale(self.rearrange(out))

    def process_dict(self, input_object, r_depth=0):
        """takes in a dict, check if values are list, if list send to process_list
        else process_primitive"""
        out = {}

        # if hasattar "templates" then the keys in input object should match
        if hasattr(self, "templates"):
            assert set(input_object.keys()) == set(
                self.templates.keys()
            ), f"input object keys do not match templates: {set(input_object.keys()) - set(self.templates.keys())}"

        for k, v in input_object.items():
            # if templates are given and the ket is same, then load that
            target_shape = None
            if hasattr(self, "templates"):
                target_shape = self.templates[k]
                target_shape = target_shape[-2:][::-1]  # [h,w] -> [w,h]

            # call the underlying object (structure or primitive)
            if isinstance(v, list):
                out[k] = self.process_list(v, target_shape, r_depth=r_depth + 1)
            elif isinstance(v, dict):
                out[k] = self.process_dict(v, r_depth=r_depth + 1)
            else:
                out[k] = self.process_primitive(v, target_shape)
        return out

    def process_list(self, input_object, target_shape=None, r_depth=0):
        """takes in a list. This function is very tricky because the input
        can be a list or a p-list, so we first check if input is not a string, Image or dict"""

        # r_depth is the depth in the current reccursion, possible depths
        # r_depth=0 -> [URL, URL]
        # r_depth=1 -> {k:[URL, URL], l:[URL, URL]}
        # r_depth=2 -> [{k:[URL, URL], l:[URL, URL]}, {k:[URL, URL], l:[URL, URL]}]}]
        # thus if r_depth > 3, raise error
        if r_depth >= 3:
            raise RecursionError("Cannot go deeper with a list input")

        if r_depth == 0 and hasattr(self, "templates") and len(self.templates) > 1:
            raise ValueError(f"Template has more than 1 input, please input a dict with keys: {tuple(self.templates.keys())}")

        if isinstance(input_object[0], (str, Image.Image)):
            logger.info(" - ImageParser - (B1) list - (str, Image.Image)")
            out = [self.process_primitive(x, target_shape) for x in input_object]
            return np.vstack(out)
        elif isinstance(input_object[0], dict):
            logger.info(" - ImageParser - (B2) list - (dict)")
            assert all([set(input_object[0].keys()) == set(i.keys()) for i in input_object]), "All keys must be same in all dicts in list"
            out = [self.process_dict(x) for x in input_object]
            out = {k: np.vstack([x[k] for x in out]) for k in out[0].keys()}
            return out
        else:
            # check if this is a list of lists or np.ndarrays
            if isinstance(input_object[0], list):
                logger.info(" - ImageParser - (B3) list - (list)")
                # convert input_object to a np.array and check shapes - used in nbox-dply
                out = np.array(input_object)
                if len(out.shape) == 3:
                    out = out[None, ...]
                return out
            else:
                logger.info(" - ImageParser - (B4) list - (primitive)")
                out = [self.process_primitive(x, target_shape) for x in input_object]
                return np.vstack(out)


class TextParser(BaseParser):
    def __init__(self, tokenizer, max_len=None, **kwargs):
        """Unified Text parsing engine, returns tokenized dictionaries

        Args:
            tokenizer (Tokenizer): tokenizer object for the text
            max_len (int): maximum length of the text
        """
        super().__init__()
        # tokenizer is supposed to be AutoTokenizer object, check that
        self.tokenizer = tokenizer
        self.max_len = max_len
        for k, v in kwargs.items():
            setattr(self, k, v)

    def process_primitive(self, x):
        # in case of text this is quite simple because only primitive is strings
        if isinstance(x, str):
            logger.info(" - TextParser - (C1) string")
            return {
                k: np.array(v)[None, ...]
                for k, v in self.tokenizer(
                    text=x,
                    add_special_tokens=True,
                    max_length=self.max_len,
                    padding="max_length" if self.max_len is not None else False,
                ).items()
            }
        elif isinstance(x, np.ndarray):
            logger.info(" - TextParser - (C2) ndarray")
            return x
        else:
            raise ValueError(f"Unsupported type for TextParser: {type(x)}")

    def process_dict(self, input_object):
        """takes in a dict and for each key's type call that method"""
        out = {}
        for k, v in input_object.items():
            if isinstance(v, list):
                out[k] = self.process_list(v)
            elif isinstance(v, dict):
                out[k] = self.process_dict(v)
            else:
                out[k] = self.process_primitive(v)
        return out

    def process_list(self, input_object):
        """takes in and tokenises the strings"""
        assert all([isinstance(x, str) for x in input_object]), "TextParser - (B1) input must be list of strings"
        return {k: np.array(v) for k, v in self.tokenizer(input_object, padding="longest").items()}



import os
import json
import logging
import requests
from getpass import getpass

from .utils import join, nbox_session, NBOX_HOME_DIR, isthere

logger = logging.getLogger()

# ------ AWS Auth ------ #

class AWSClient:
  @isthere("boto3", "botocore", soft = False)
  def __init__(self, aws_access_key_id, aws_secret_access_key, region_name):
    self.aws_access_key_id = aws_access_key_id
    self.aws_secret_access_key = aws_secret_access_key
    self.region_name = region_name

  def get_client(self, service_name = "s3", **boto_config_kwargs):
    import boto3
    from botocore.client import Config as BotoConfig

    return boto3.client(
      service_name,
      aws_access_key_id=self.aws_access_key_id,
      aws_secret_access_key=self.aws_secret_access_key,
      region_name=self.region_name,
      config = BotoConfig(
        signature_version="s3v4",
        **boto_config_kwargs
      )
    )

# ------ GCP Auth ------ #

class GCPClient:
  @isthere("google-cloud-sdk", "google-cloud-storage", soft = False)
  def __init__(self, project_id, credentials_file):
    from google.oauth2 import service_account
    
    self.project_id = project_id
    self.credentials_file = credentials_file
    self.creds = service_account.Credentials.from_service_account_file(
      self.credentials_file,
      scopes=["https://www.googleapis.com/auth/cloud-platform"]
    )

  def get_client(self, service_name = "storage", **gcp_config_kwargs):
    if service_name == "storage":
      from google.cloud import storage
      return storage.Client(
        project=self.project_id,
        credentials=self.creds,
        **gcp_config_kwargs
      )
    

# ------ Azure Auth ------ #

class AzureClient:
  @isthere("azure-storage-blob", soft = False)
  def __init__(self):
    from azure.storage.blob import BlobServiceClient
    from azure.identity import DefaultAzureCredential

    self.blob_service_client = BlobServiceClient(
      credential=DefaultAzureCredential(),
      endpoint="https://nbox.blob.core.windows.net"
    )

  def get_client(self, service_name = "blob", **azure_config_kwargs):
    if service_name == "blob":
      from azure.storage.blob import BlobClient

      return BlobClient(
        self.blob_service_client,
        **azure_config_kwargs
      )

# ------ OCI Auth ------ #

class OCIClient:
  @isthere("oci", "oci-py", soft = False)
  def __init__(self, config_file):
    from oci.config import from_file
    from oci.signer import Signer

    self.config = from_file(config_file)
    self.signer = Signer(
      tenancy=self.config["tenancy"],
      user=self.config["user"],
      fingerprint=self.config["fingerprint"],
      private_key_file_location=self.config["key_file"]
    )

  def get_client(self, service_name = "object_storage", **oci_config_kwargs):

    if service_name == "object_storage":
      from oci.object_storage.models import CreateBucketDetails
      from oci.object_storage.models import CreateMultipartUploadDetails
      from oci.object_storage.models import Object
      from oci.object_storage.models import UploadPartDetails
      from oci.object_storage.object_storage_client import ObjectStorageClient

      return ObjectStorageClient(
        self.config["user"],
        self.signer,
        **oci_config_kwargs
      )


# ------ Digital Ocean Auth ------ #

class DOClient:
  @isthere("doctl", soft = False)
  def __init__(self, config_file):
    from doctl.doctl_client import DictCursor
    from doctl.doctl_client import DoctlClient

    self.doctl_client = DoctlClient(
      config_file=config_file,
      cursor_class=DictCursor
    )
  
  def get_client(self, service_name = "object_storage", **oci_config_kwargs):
    if service_name == "object_storage":
      from doctl.object_storage.object_storage_client import ObjectStorageClient

      return ObjectStorageClient(
        self.doctl_client,
        **oci_config_kwargs
      )

# ------ NBX Auth ------ #

class NBXClient:
  @staticmethod
  def get_access_token(nbx_home_url, username, password=None):
    password = getpass("Password: ") if password is None else password
    try:
      r = requests.post(url=f"{nbx_home_url}/api/user/login", json={"username": username, "password": password})
    except Exception as e:
      raise Exception(f"Could not connect to NBX | {str(e)}")

    if r.status_code == 401:
      logger.error(" Invalid username/password. Please try again!")
      return False
    elif r.status_code == 200:
      access_packet = r.json()
      access_token = access_packet.get("access_token", None)
      logger.info("Access token obtained")
      return access_token
    else:
      logger.error(f"Unknown error: {r.content.decode()}")
      raise Exception(f"Unknown error: {r.status_code}")

  @staticmethod
  def create_secret_file(username, access_token, nbx_url):
    os.makedirs(NBOX_HOME_DIR, exist_ok=True)
    fp = join(NBOX_HOME_DIR, "secrets.json")
    with open(fp, "w") as f:
      f.write(json.dumps({"username": username, "access_token": access_token, "nbx_url": nbx_url}, indent=2))

  def __init__(self):
    os.makedirs(NBOX_HOME_DIR, exist_ok=True)
    fp = join(NBOX_HOME_DIR, "secrets.json")

    # if this is the first time starting this then get things from the nbx-hq
    if not os.path.exists(fp):
      # get the secrets JSON
      try:
        self.secrets = json.loads(requests.get(
          "https://raw.githubusercontent.com/NimbleBoxAI/nbox/master/assets/sample_config.json"
          ).content.decode("utf-8")
        )
      except Exception as e:
        raise Exception(f"Could not create secrets file: {e}")

      # populate with the first time things
      nbx_home_url = "https://www.nimblebox.ai"
      username = input("Username: ")
      access_token = None
      while not access_token:
        access_token = self.get_access_token(nbx_home_url, username)
        self.secrets["access_token"] = access_token
      self.secrets["username"] = username
      self.secrets["nbx_url"] = nbx_home_url
      with open(fp, "w") as f:
        f.write(self.__repr__())
      logger.info("Successfully created secrets!")
    else:
      with open(fp, "r") as f:
        self.secrets = json.load(f)
      logger.info("Successfully loaded secrets!")

  def __repr__(self):
    return json.dumps(self.secrets)

  def get(self, item):
    return self.secrets[item]

# function for manual trigger

def init_secret():
  global secret
  secret = NBXClient()
  nbox_session.headers.update({"Authorization": f"Bearer {secret.get('access_token')}"})

secret = None
if not os.getenv("NBOX_NO_AUTH", False):
  init_secret()



import os
import logging

def reset_log():
  JSON_LOG = os.environ.get("NBOX_JSON_LOG", False)
  json_format = '{"time": "%(asctime)s", "level": "%(levelname)s", "message": "%(message)s"}'
  normal_format = '[%(asctime)s] [%(levelname)s] %(message)s'

  logging.basicConfig(
    level = logging.INFO,
    format = json_format if JSON_LOG else normal_format,
    datefmt = "%Y-%m-%dT%H:%M:%S%z" # isoformat
  )



import json
import sys
import os
from typing import List
from time import sleep

from .network import deploy_model
from .auth import init_secret, secret
from .utils import get_random_name, NBOX_HOME_DIR, join
from .jobs import get_instance, Instance

def status(loc = None):
    from .jobs import print_status
    print_status(f"https://{'' if not loc else loc+'.'}nimblebox.ai")


def tunnel(ssh: int, *apps_to_ports: List[str], i: str):
    """the nbox way to SSH into your instance, by default ``"jupyter": 8888 and "mlflow": 5000``

    Usage:
        tunn.py 8000 notebook:8000 2000:8001 -i "nbox-dev"

    Args:
        ssh: Local port to connect SSH to
        *apps_to_ports: A tuple of values ``<app_name/instance_port>:<localport>``.
            For example, ``jupyter:8888`` or ``2001:8002``
        i(str): The instance to connect to
        pwd (str): password to connect to that instance.
    """

    import socket
    import logging

    import ssl
    import threading
    import certifi

    log_ = open(join(NBOX_HOME_DIR, "tunnel.log"), "w")
    logger = lambda x: log_.write(x + "\n")

    class RSockClient:
        """
        This is a RSockClient. It handels the client socket where client is the user application trying to connect to "client_port"
        Connects to RSockServer listening on localhost:886.
        RSockServer recieves instructions as a string and sends back a response.
        RSockServer requires following steps to setup
        First,
            Authentication:
                - Authentication happens by sending
                    `"AUTH~{AUTH_TOKEN}"`
                - AUTH_TOKEN is not defined and is default to 'password'
            Seting config:
                - you can set config by sending
                    `"SET_CONFIG~{instance}~{instance_port}"`
                - "instance" - Currently is the internal ip of the instance.
                - "instance_port" - What port users wants to connect to.
            Connect:
                - This Starts the main loop which
                    1. Listen on client_port
                    2. On connection, 
                        a. Send AUTH
                        b. If AUTH is successful, send SET_CONFIG
                        c. If SET_CONFIG is successful, send CONNECT
                        d. If CONNECT is successful, start io_copy
            IO_COPY:
                - This is the main loop that handles the data transfer between client and server. This is done by creating a new thread for each connection.
                - The thread is created by calling the function "io_copy" for each connection that is "server" and "client".
                - When a connection is closed, the loop is stopped.
        """

        def __init__(self, connection_id, client_socket, instance, instance_port, auth, secure=False):
            """
            Initializes the client.
            Args:
                client_socket: The socket that the client is connected to.
                instance: The instance that the client wants to connect to.
                instance_port: The port that the instance is listening on.
                auth: The authentication token that the client has to provide to connect to the RSockServer.
                secure: Whether or not the client is using SSL.
            
            """
            self.connection_id = connection_id
            self.client_socket = client_socket
            self.instance = instance
            self.instance_port = instance_port
            self.auth = auth
            self.secure = secure

            self.client_auth = False
            self.rsock_thread_running = False
            self.client_thread_running = False


            self.log('Starting client')
            self.connect_to_rsock_server()
            self.log('Connected to RSockServer')
            self.authenticate()
            self.log('Authenticated client')
            self.set_config()
            self.log('Client init complete')
        
        def log(self, message, level=logging.INFO):
            logger(f"[Client_ID: {self.connection_id}] {message}")
            # print(f"[Client_ID: {self.connection_id}] {message}")
        
        def connect_to_rsock_server(self):
            """
            Connects to RSockServer.
            """
            self.log('Connecting to RSockServer', logging.DEBUG)
            rsock_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            rsock_socket.connect(('rsocks.nimblebox.ai', 886))

            if self.secure:
                self.log('Starting SSL')
                self.rsock_socket = ssl.wrap_socket(rsock_socket, ca_certs=certifi.where(), cert_reqs=ssl.CERT_REQUIRED)
            else:
                self.rsock_socket = rsock_socket

        def authenticate(self):
            """
            Authenticates the client.
            Sends `"AUTH~{AUTH_TOKEN}"` to RSockServer.
            """
            self.log('Authenticating client')
            self.rsock_socket.sendall(bytes('AUTH~{}'.format(self.auth), 'utf-8'))
            auth = self.rsock_socket.recv(1024)
            auth = auth.decode('utf-8')
            if auth == 'OK':
                self.log('Client authenticated')
            else:
                self.log('Client authentication failed', logging.ERROR)
                self.client_auth = False
                exit(1)
        
        def set_config(self):
            """
            Sets the config of the client.
            Sends `"SET_CONFIG~{instance}~{instance_port}"` to RSockServer.
            """
            self.log('Setting config')
            self.rsock_socket.sendall(bytes(f'SET_CLIENT~{self.instance}~{self.instance_port}', 'utf-8'))
            config = self.rsock_socket.recv(1024)
            config = config.decode('utf-8')
            self.log('Config set to {}'.format(config))
            if config == 'OK':
                self.log('Config set')
            else:
                self.log('Config set failed', logging.ERROR)
                exit(1)

        def connect(self):
            """
            Connects the client to RSockServer.
            Sends `"CONNECT"` to RSockServer.
            """
            self.log('Starting the io_copy loop')
            self.rsock_socket.sendall(bytes('CONNECT', 'utf-8'))
            
            # start the io_copy loop
            self.rsock_thread_running = True
            self.rsock_thread = threading.Thread(target=self.io_copy, args=("server", ))
            self.rsock_thread.start()

            # start the io_copy loop
            self.client_thread_running = True
            self.client_thread = threading.Thread(target=self.io_copy, args=("client", ))
            self.client_thread.start()

        def io_copy(self, direction):
            """
            This is the main loop that handles the data transfer between client and server.
            """
            self.log('Starting {} io_copy'.format(direction))

            if direction == 'client':
                client_socket = self.client_socket
                server_socket = self.rsock_socket

            elif direction == 'server':
                client_socket = self.rsock_socket
                server_socket = self.client_socket

            while True:
                try:
                    data = client_socket.recv(1024)
                    if data:
                        # self.log('{} data: {}'.format(direction, data))
                        server_socket.sendall(data)
                    else:
                        self.log('{} connection closed'.format(direction))
                        break
                except Exception as e:
                    self.log('Error in {} io_copy: {}'.format(direction, e), logging.ERROR)
                    break
            self.log('Stopping {} io_copy'.format(direction))


    def create_connection(local_port, instance_id, instance_port, listen = 1):
        listen_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        listen_socket.bind(('localhost', local_port))
        listen_socket.listen(listen)

        connection_id = 0

        while True:
            logger('Waiting for client')
            client_socket, _ = listen_socket.accept()
            logger('Client connected')

            connection_id += 1
            logger('Total clients connected -> '.format(connection_id))
            # create the client
            pwd = secret.get("access_token")

            client = RSockClient(connection_id, client_socket, instance_id, instance_port, pwd, True)

            # start the client
            client.connect()

    def port_in_use(port: int) -> bool:
        import socket
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex(('localhost', port)) == 0
    
    # ===============

    default_ports = {
        "jupyter": 8888,
        "mlflow": 5000,
    }
    apps = {} # <localport-cloudport>
    for ap in apps_to_ports:
        app, port = ap.split(':')
        port = int(port)
        if not app or not port:
            raise ValueError(f"Invalid app:port pair {ap}")
        try:
            apps[int(app)] = port
        except:
            if app not in default_ports:
                raise ValueError(f"Unknown '{app}' should be either integer or one of {', ' .join(default_ports.keys())}")
            apps[port] = default_ports[app]
    apps[ssh] = 22 # hard code

    ports_used = []
    for k in apps:
        if port_in_use(k):
            ports_used.append(str(k))

    if ports_used:
        raise ValueError(f"Ports {', '.join(ports_used)} are already in use")

    # check if instance is the correct one
    instance = Instance(i, loc = "test-3")
    if not instance.state == "RUNNING":
        raise ValueError("Instance is not running")
    passwd = instance.open_data["ssh_pass"]
    logging.info(f"password: {passwd}")

    # create the connection
    threads = []
    for local_port, cloud_port in apps.items():
        logging.info(f"Creating connection from {cloud_port} -> {local_port}")
        t = threading.Thread(target=create_connection, args=(local_port, instance.instance_id, cloud_port, 1))
        t.start()
        threads.append(t)

    try:
        # start the ssh connection on terminal
        import subprocess
        logging.info(f"Starting SSH ... for graceful exit press Ctrl+D then Ctrl+C")
        subprocess.call(f'ssh -p {ssh} ubuntu@localhost', shell=True)
    except KeyboardInterrupt:
        logging.info("KeyboardInterrupt, closing connections")
        for t in threads:
            t.join()

    # TODO:@yashbonde Make Platform agnostic
    subprocess.run(["ssh-keygen", "-R", f"localhost[{ssh}]"])
    sys.exit(0) # graceful exit


def deploy(
    config_path: str = None,
    model_path: str = None,
    model_name: str = None,
    nbox_meta: str = None,
    deployment_type: str = None,
    deployment_id: str = None,
    deployment_name: str = None,
    convert_args: str = None,
    wait_for_deployment: bool = False,
    print_in_logs: bool = False,
    username: str = None,
    password: str = None,
    nbx_home_url="https://nimblebox.ai",
):
    r"""Deploy a model from nbox CLI. Add this to your actions and see the magic happen!
    If you are using a config file then data will be loaded from there and other kwargs will be ignored.

    Args:
        config_path (str, optional): path to your config file, if provided, everything else is ignored
        model_path (str, optional): path to your model
        model_name (str, optional): name of your model
        nbox_meta ([str, dict], optional): path to your nbox_meta json file, if None tries to find by replacing
          ``model_path`` extension with ``.json`` or can be a dict if using ``config_path``
        deployment_type (str, optional): type of deployment, can be one of: ovms2, nbox
        convert_args (str, optional): if using ovms2 deployment type, you must pass convertion CLI args
        wait_for_deployment (bool, optional): wait for deployment to finish, if False this behaves async
        print_in_logs (bool, optional): print logs in stdout
        username (str, optional): your NimbleBox.ai username
        password (str, optional): your password for corresponding ``username``

    Usage:
        Convert each ``kwarg`` to ``--kwarg`` for CLI. eg. if you want to pass value for ``model_path`` \
            in cli it becomes like ``... --model_path="my_model_path" ...``

        .. code-block:: bash

            python3 -m nbox deploy --model_path="path/some/sklearn.pkl" \
                --nbox_meta="path/to/nbox_meta.json"

    Raises:
        ValueError: if ``deployment_type`` is not supported
        AssertionError: if model path is not found or ``nbox_meta`` is incorrect
        Exception: if ``deployment_type == "ovms2"`` but ``convert_args`` is not provided
    """
    from nbox.auth import secret  # it can refresh so add it in the method

    if secret is None or secret.get("access_token", None) == None:
        # if secrets file is not found
        assert username != None and password != None, "secrets.json not found need to provide username password for auth"
        access_token = get_access_token(nbx_home_url, username, password)
        create_secret_file(username, access_token, nbx_home_url)
        init_secret()  # reintialize secret variable as it will be used everywhere

    if config_path != None:
        with open(config_path, "r") as f:
            config = json.load(f)
        config.pop("config_path", None) # remove recursion
        deploy(**config)

    else:
        # is model path valid and given
        if not os.path.exists(model_path):
            assert os.path.exists(model_path), "model path not found"

        # check if nbox_meta is correct
        if nbox_meta == None:
            nbox_meta = ".".join(model_path.split(".")[:-1]) + ".json"
            print("Trying to find nbox meta at path:", nbox_meta)
            assert os.path.exists(nbox_meta), "nbox_meta not provided"
        else:
            raise ValueError("nbox_meta is not supported yet")

        if isinstance(nbox_meta, str):
            if not os.path.exists(nbox_meta):
                raise ValueError(f"Nbox meta path {nbox_meta} does not exist. see nbox.Model.get_nbox_meta()")
            with open(nbox_meta, "r") as f:
                nbox_meta = json.load(f)
        else:
            assert isinstance(nbox_meta, dict), "nbox_meta must be a dict"

        # validation of deployment_type
        assert deployment_type in ("ovms2", "nbox"), "Deployment type must be one of: ovms2, nbox"
        if deployment_type == "ovms2":
            assert convert_args is not None, (
                "Please provide convert args when using OVMS deployment, "
                "use nbox.Model.deploy(deployment_type == 'ovms2') if you are unsure!"
            )

        # one click deploy
        model_name = get_random_name().replace("-", "_") if model_name == None else model_name
        endpoint, key = deploy_model(
            model_path, model_name, deployment_type, nbox_meta, wait_for_deployment, convert_args, deployment_id, deployment_name
        )

        # print to logs if needed
        if wait_for_deployment and print_in_logs:
            print(" Endpoint:", endpoint)
            print("Model Key:", key)



# Some parts of the code are based on the pytorch nn.Module class
# pytorch license: https://github.com/pytorch/pytorch/blob/master/LICENSE
# due to requirements of stability, some type enforcing is performed

import os
from typing import Callable
from functools import partial
from tempfile import gettempdir
from collections import OrderedDict
from datetime import datetime, timedelta

from logging import getLogger
logger = getLogger()

from ..network import deploy_job
from ..utils import join
from ..framework.__airflow import AirflowMixin
from ..framework.on_functions import get_nbx_flow, DBase


class StateDictModel(DBase):
  __slots__ = [
    "state", # :str
    "data", # :dict
    "inputs", # :dict
    "outputs", # :dict
  ]


# class TraceObject:
#   def __init__(self, root):
#     self.root = root
#     self.flow = OrderedDict()
#
#   def pre(self, inputs, cls):
#     _id = id(cls)
#     self.flow[_id] = {
#       "id": _id,
#       "class_name": cls.__class__.__name__,
#       "inputs": {k: type(v) for k, v in inputs.items()},
#       "outputs": {},
#       "start": datetime.now(),
#       "end": None,
#     }
#
#   def post(self, out, cls):
#     _id = id(cls)
#     if _id not in self.flow:
#       raise ValueError(f"{_id} not found in flow")
#
#     outputs = {}
#     if out == None:
#       outputs = {"out_0": type(None)}
#     elif isinstance(out, dict):
#       outputs = {k: type(v) for k, v in out.items()}
#     elif isinstance(out, (list, tuple)):
#       outputs = {f"out_{i}": type(v) for i, v in enumerate(out)}
#     else:
#       outputs = {"out_0": type(out)}
#
#     self.flow[_id].update({
#       "outputs": outputs,
#       "end": datetime.now(),
#     })
#     self.flow[_id]["duration"] = self.flow[_id]["end"] - self.flow[_id]["start"]
#
#     # convert datetime to string objects for serialization
#     self.flow[_id]["start"] = self.flow[_id]["start"].isoformat()
#     self.flow[_id]["end"] = self.flow[_id]["end"].isoformat()
#     self.flow[_id]["duration"] = str(self.flow[_id]["duration"].total_seconds())
#
#   def to_dict(self):
#     return {
#       "root": self.root.__class__.__name__,
#       "root_id": id(self.root),
#       "flow": self.flow,
#     }
#
#   def dag(self, depth = 1, root_ = None):
#     if depth != 1:
#       raise ValueError("depth of 1 supported only")
#
#     if depth < 0:
#       return []
#
#     dag = []
#     # create nodes
#     root_ = root_ if root_ != None else self.root
#     for _name, c in root_._operators.items():
#       _id = id(c)
#       name = c.__class__.__name__
#       _trace = self.flow[_id]
#       _trace["code_name"] = _name
#
#       # if there is some depth left, recurse
#       if depth > 1:
#         children_dag = self.dag(depth - 1, c)
#         _trace["children"] = children_dag
#
#       dag.append({
#         "id": _id,
#         "type": "input" if len(dag) == 0 else None,
#         "data": {
#           "label": f"{_name} | {name}"
#         },
#         # "meta": _trace
#       })
#     dag[-1]["type"] = "output"
#
#     # create edges
#     for src, trg in zip(dag[:-1], dag[1:]):
#       dag.append({
#         "id": f"edge-{src['id']}-{trg['id']}",
#         "source": src["id"],
#         "target": trg["id"]
#       })
#   
#     return dag

class Tracer:
  def __init__(self):
    try:
      # when job is running on NBX, gRPC stubs are used
      import nbox_js_stub
      self.l = nbox_js_stub.Trace()
      self._trace_obj = "stub"
    except ImportError:
      def _trace(x, fn):
        logger.info(x, extra={"fn": fn})

      self.l = _trace
      self._trace_obj = "logger"

  def __getattr__(self, __name):
    item = getattr(self.l, __name, None)
    if self._trace_obj == "logger":
      return partial(self.l, fn=__name)
    elif self._trace_obj == "stub" and item:
      return item
    raise AttributeError(f"Service: '{__name}' not found")

  def __call__(self, dag_update):
    # import requests
    # r = requests.post(
    #   url = "127.0.0.1:8000/log",
    #   json = dag_update,
    # )
    # print(r.content)
    print(dag_update)


class Operator(AirflowMixin):
  _version: int = 1

  def __init__(self) -> None:
    self._operators = OrderedDict() # {name: operator}
    self._op_trace = []

  # classmethods/

  def serialise(self):
    pass

  @classmethod
  def deserialise(cls, state_dict):
    pass

  # AirflowMixin methods
  # --------------------
  # AirflowMixin.to_airflow_operator(self, timeout, **operator_kwargs):
  # AirflowMixin.from_airflow_operator(cls, air_operator) <- classmethod
  # AirflowMixin.to_airflow_dag(self, dag_kwargs, operator_kwargs)
  # AirflowMixin.from_airflow_dag(cls, dag) <- classmethod

  # /classmethods

  def __repr__(self):
    # from torch.nn.Module
    def _addindent(s_, numSpaces):
      s = s_.split('\n')
      # don't do anything for single-line stuff
      if len(s) == 1:
        return s_
      first = s.pop(0)
      s = [(numSpaces * ' ') + line for line in s]
      s = '\n'.join(s)
      s = first + '\n' + s
      return s

    # We treat the extra repr like the sub-module, one item per line
    extra_lines = []
    extra_repr = ""
    # empty string will be split into list ['']
    if extra_repr:
      extra_lines = extra_repr.split('\n')
    child_lines = []
    for key, module in self._operators.items():
      mod_str = repr(module)
      mod_str = _addindent(mod_str, 2)
      child_lines.append('(' + key + '): ' + mod_str)
    lines = extra_lines + child_lines

    main_str = self.__class__.__name__ + '('
    if lines:
      # simple one-liner info, which most builtin Modules will use
      if len(extra_lines) == 1 and not child_lines:
        main_str += extra_lines[0]
      else:
        main_str += '\n  ' + '\n  '.join(lines) + '\n'

    main_str += ')'
    return main_str

  def __setattr__(self, key, value: 'Operator'):
    obj = getattr(self, key, None)
    if key != "forward" and obj is not None and callable(obj):
      raise AttributeError(f"cannot assign {key} as it is already a method")
    if isinstance(value, Operator):
      if not "_operators" in self.__dict__:
        raise AttributeError("cannot assign operator before Operator.__init__() call")
      if key in self.__dict__ and key not in self._operators:
        raise KeyError(f"attribute '{key}' already exists")
      self._operators[key] = value
    self.__dict__[key] = value

  # properties/

  def operators(self):
    r"""Returns an iterator over all operators in the job."""
    for _, module in self.named_operators():
      yield module

  def named_operators(self, memo = None, prefix: str = '', remove_duplicate: bool = True):
    r"""Returns an iterator over all modules in the network, yielding
    both the name of the module as well as the module itself."""
    if memo is None:
      memo = set()
    if self not in memo:
      if remove_duplicate:
        memo.add(self)
      yield prefix, self
      for name, module in self._operators.items():
        if module is None:
          continue
        submodule_prefix = prefix + ('.' if prefix else '') + name
        for m in module.named_operators(memo, submodule_prefix, remove_duplicate):
          yield m

  @property
  def children(self):
    return self._operators.values()

  # user can manually define inputs if they want, avoid this user if you don't know
  # how this system works
  _inputs = []
  @property
  def inputs(self):
    import inspect
    args = inspect.getfullargspec(self.forward).args
    try:
      args.remove('self')
    except:
      pass
    if self._inputs:
      args += self._inputs
    return args

  @property
  def state_dict(self) -> StateDictModel:
    return StateDictModel(
      state = self.__class__.__name__,
      data = {},
      inputs = self.inputs,
      outputs = self.outputs
    )

  # properties as planned

  @property
  def comms(self):
    # this returns all the things communications such as email, Slack, Discord
    # phone alerts, system notifications, gmeets, basically everything supported
    # in https://github.com/huggingface/knockknock/tree/master/knockknock
    raise NotImplementedError()

  # /properties

  # information passing/

  def propagate(self, **kwargs):
    for c in self.children:
      c.propagate(**kwargs)
    for k, v in kwargs.items():
      setattr(self, k, v)

  # /information passing

  def forward(self):
    raise NotImplementedError("User must implement forward()")

  def _register_forward(self, python_callable: Callable):
    # convienience method to register a forward method
    self.forward = python_callable

  _trace_object = Tracer()
  node_info = None
  source_edges = None

  # def trace(self, *args, return_dag = True, **kwargs):
  #   self._trace_object = TraceObject(self)
  #   self.propagate(_trace_object = self._trace_object)
  #   self(*args, **kwargs)
  #   trace = self._trace_object.to_dict()
  #   dag = self._trace_object.dag()
  #   self.propagate(_trace_object = None)
  #   self._trace_object = None
  #
  #   output = (trace,)
  #   if return_dag:
  #     output += (dag,)
  #   return output

  def __call__(self, *args, **kwargs):
    # blank comment so docstring below is not loaded

    """There is no need to perform ``self.is_dag`` check here, since it is
    a declarative model not an imperative one, so existance of DAG's is
    irrelevant.

    This function will has the code for the following:
    1. Input checks, we want to enforce that
    2. Tracing, we want to trace the execution of the model
    3. Networking, when required to execute like RPC
    """

    # Type Checking and create input dicts
    inputs = self.inputs
    len_inputs = len(args) + len(kwargs)
    if len_inputs > len(inputs):
      raise ValueError(f"Number of arguments ({len(inputs)}) does not match number of inputs ({len_inputs})")
    elif len_inputs < len(args):
      raise ValueError(f"Need at least arguments ({len(args)}) but got ({len_inputs})")

    input_dict = {}
    for i, arg in enumerate(args):
      input_dict[self.inputs[i]] = arg
    for key, value in kwargs.items():
      if key in inputs:
        input_dict[key] = value

    if self.node_info != None:
      self.node_info["run_status"]["start"] = datetime.now().isoformat()
      self.node_info["run_status"]["inputs"] = {k: str(type(v)) for k, v in input_dict.items()}
      self._trace_object(self.node_info)

    # ----input
    out = self.forward(**input_dict) # pass this through the user defined forward()
    # ----output

    if self.node_info != None:
      outputs = {}
      if out == None:
        outputs = {"out_0": type(None)}
      elif isinstance(out, dict):
        outputs = {k: type(v) for k, v in out.items()}
      elif isinstance(out, (list, tuple)):
        outputs = {f"out_{i}": type(v) for i, v in enumerate(out)}
      else:
        outputs = {"out_0": type(out)}
      self.node_info["run_status"]["end"] = datetime.now().isoformat()
      self.node_info["run_status"]["outputs"] = outputs
      self._trace_object(self.node_info)

    return out

  # nbx/

  def thaw(self, flowchart):
    nodes = flowchart["nodes"]
    edges = flowchart["edges"]
    for n in nodes:
      name = n["name"]
      if name.startswith("self."):
        name = name[5:]
      if hasattr(self, name):
        op = getattr(self, name)
        op.propagate(
          node_info = n,
          source_edges = list(filter(
            lambda x: x["target"] == n["id"], edges
          ))
        )

  def deploy(
    self,
    init_folder: str = None,
    cache_dir: str = None,
    job_id = None,
    job_name = None,
    start_datetime: datetime = None,
    end_datetime: datetime = None,
    time_interval: timedelta = None,
  ):
    logger.info(f"Deploying {self.__class__.__name__} -> '{job_id}/{job_name}'")

    # flowchart-alpha
    dag = get_nbx_flow(self.forward)
    try:
      from json import dumps
      dumps(dag)
    except:
      logger.error("Cannot perform pre-building, only live updates will be available!")
      logger.info("Please raise an issue on chat to get this fixed")
      dag = {"flowchart": None, "symbols": None}

    for n in dag["flowchart"]["nodes"]:
      name = n["name"]
      if name.startswith("self."):
        name = name[5:]
      operator_name = "CodeBlock" # default
      cls_item = getattr(self, name, None)
      if cls_item and cls_item.__class__.__base__ == Operator:
        operator_name = cls_item.__class__.__name__
      n["operator_name"] = operator_name

    schedule_meta = {
      "start_datetime": start_datetime,
      "end_datetime": end_datetime,
      "time_interval": time_interval,
      "job_id": job_id,
      "job_name": job_name,
      "dag": dag,
      "created": datetime.now().isoformat(),
    }
    # print(schedule_meta)
    return schedule_meta

    # check if this is a valif folder or not
    if not os.path.exists(init_folder) or not os.path.isdir(init_folder):
      raise ValueError(f"Incorrect project at path: '{init_folder}'! nbox jobs init <name>")
    if os.path.isdir(init_folder):
      os.chdir(init_folder)
      if not os.path.exists("./exe.py"):
        raise ValueError(f"Incorrect project at path: '{init_folder}'! nbox jobs init <name>")
    else:
      raise ValueError(f"Incorrect project at path: '{init_folder}'! nbox jobs init <name>")

    # zip the folder
    import zipfile
    zip_path = join(cache_dir if cache_dir else gettempdir(), "project.zip")
    logger.info(f"Zipping project to '{zip_path}'")
    zip_file = zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED)
    for root, dirs, files in os.walk(init_folder):
      for file in files:
        zip_file.write(os.path.join(root, file))
    zip_file.close()

    return schedule_meta

    # deploy_job(zip_path = zip_path, schedule_meta = schedule_meta)

  # /nbx



# this file has the code for nbox.Model that is the holy grail of the project

import os
import json
import shutil
import inspect
import tarfile
import requests
from glob import glob
from time import time
from tempfile import gettempdir
from pprint import pprint as pp

# import nbox things
from . import utils

# required for lazy loading --> the underlying framework as the information
# on what modules are to be loaded using IMPORTS variable defined in each framework
# conditional
from .framework import get_meta, __all__ as framework_exports
for x in framework_exports[1:]:
    exec(f"from .framework import {x} as frm_{x}")
    IMPORTS = eval(f"frm_{x}.IMPORTS")
    for y in IMPORTS:
        exec(f"import {y}")

from .network import deploy_model
from .parsers import ImageParser, TextParser
from .auth import secret

import logging

logger = logging.getLogger()


class GenericMixin:
    def eval(self):
        """if underlying model has eval method, call it"""
        if hasattr(self.model_or_model_url, "eval"):
            self.model_or_model_url.eval()

    def train(self):
        """if underlying model has train method, call it"""
        if hasattr(self.model_or_model_url, "train"):
            self.model_or_model_url.train()

    def __repr__(self):
        return f"<nbox.Model: {self.model_or_model_url} >"


class Model(GenericMixin):
    # @cshubhamrao suggested lazy evaulation of packages that would avoid initialization Errors
    @utils.isthere("torch", "sklearn", "onnxruntime", "skl2onnx")
    def __init__(self, model_or_model_url, nbx_api_key=None, category=None, tokenizer=None, model_key=None, nbox_meta=None, cache_dir = None, verbose=False):
        """Model class designed for inference. Seemlessly remove boundaries between local and cloud inference
        from ``nbox==0.1.10`` ``nbox.Model`` handles both local and remote models. from ``nbox==0.4.0`` can load
        and serve ONNX models.

        Usage:

            .. code-block:: python

                from nbox import Model

                # when on NBX-Deploy
                model = Model("https://nbx.cloud/model/url", "nbx_api_key")

                # when loading a scikit learn model
                from sklearn.datasets import load_iris
                from sklearn.ensemble import RandomForestClassifier
                iris = load_iris()
                clr = RandomForestClassifier()
                clr.fit(iris.data, iris.target)
                model = nbox.Model(clr)

                # when loading a pytorch model
                import torch

                class DoubleInSingleOut(torch.nn.Module):
                    def __init__(self):
                        super().__init__()
                        self.f1 = torch.nn.Linear(2, 4)
                        self.f2 = torch.nn.Linear(2, 4)
                        self.logit_scale = torch.nn.Parameter(torch.ones([]) * np.log(1 / 0.07))

                    def forward(self, x, y):
                        out = self.f1(x) + self.f2(y)
                        logit_scale = self.logit_scale.exp()
                        out = logit_scale - out @ out.t()
                        return out

                model = nbox.Model(
                    DoubleInSingleOut(),
                    category = {"x": "image", "y": "image"} # <- this is pre-proc type for input
                )
                # this is not the best approach, but it works for now


        Args:
            model_or_model_url (Any): Model to be wrapped or model url
            nbx_api_key (str, optional): API key for this deployed model
            category (str, optional): Input categories for each input type to the model
            tokenizer ([transformers.PreTrainedTokenizer], optional): If this is a text model then tokenizer for this. Defaults to None.
            model_key (str, optional): With what key is this model initialised, useful for public models. Defaults to None.
            model_meta (dict, optional): Extra metadata when starting the model. Defaults to None.
            verbose (bool, optional): If true provides detailed prints. Defaults to False.

        Raises:
            ValueError: If any category is "text" and tokenizer is not provided.

        This class can process the following types of models:

        .. code-block::

            ``torch.nn.Module`` |           |           |
                    ``sklearn`` | __init__  > serialise > S-*
                 ``nbx-deploy`` |___________|___________|____
                     ``S-onnx`` |              |
              ``S-torchscript`` | .deserialise > __init__
                      ``S-pkl`` |______________|___________

        Serialised models are 1-3 models that have been serialised to load later. This is especially useful for
        ``nbox-serving``, one of the server types we use in NBX Deploy, yes production. This is also part of
        YoCo. Since there is a decidate serialise function, we should have one for deserialisation as well. Use
        ``nbox.Model.desirialise`` to load a serialised model.
        """
        self.model_or_model_url = model_or_model_url
        self.nbx_api_key = nbx_api_key
        self.verbose = verbose

        # here we have the main conditionals for loading the models
        # from code perspective avoid "self" assignment made inside each block,
        # but rather at the end of the blocks
        __framework = None
        nbox_meta = nbox_meta
        image_parser = None
        text_parser = None

        # the tricky part about lazy loading is that it needs to be done in the right order
        # so errors are seen by developer only when needed. So first we check if it is
        # a) deployed model or not
        # b) is a sklearn model
        # else: # this has to be run in try catch blocks
        #  c) is a torch model
        #  d) is a onnx model

        if isinstance(model_or_model_url, str) and model_or_model_url.startswith("http"):
            logger.info(f"Trying to load from url: {model_or_model_url}")
            __framework = "nbx"

            assert isinstance(nbx_api_key, str), "Nbx API key must be a string"
            assert nbx_api_key.startswith("nbxdeploy_"), "Not a valid NBX Api key, please check again."
            assert model_or_model_url.startswith("http"), "Are you sure this is a valid URL?"

            self.model_url = model_or_model_url.rstrip("/")
            nbox_meta, templates = self.fetch_meta_from_nbx_cloud()
            category = nbox_meta["spec"]["category"]

            # if category is "text" or if it is dict then any key is "text"
            tokenizer = None
            max_len = None
            if category == "text" or (isinstance(category, dict) and any([x == "text" for x in category.values()])):
                import transformers

                model_key = nbox_meta["spec"]["model_key"].split("::")[0].split("transformers/")[-1]
                tokenizer = transformers.AutoTokenizer.from_pretrained(model_key)
                max_len = templates["input_ids"][-1]

            image_parser = ImageParser(cloud_infer=True, post_proc_fn=lambda x: x.tolist(), templates=templates)
            text_parser = TextParser(tokenizer=tokenizer, max_len=max_len, post_proc_fn=lambda x: x.tolist())

            self.templates = templates

        elif "sklearn" in str(type(model_or_model_url)):
            logger.info(f"Trying to load from sklearn model: {model_or_model_url}")
            __framework = "sklearn"

        else:
            done = False
            try:
                if isinstance(model_or_model_url, torch.nn.Module):
                    logger.info(f"Trying to load from torch model")
                    __framework = "pytorch"
                    assert category is not None, "Category for inputs must be provided, when loading model manually"

                    model_key = model_key

                    # initialise all the parsers
                    image_parser = ImageParser(post_proc_fn=lambda x: torch.from_numpy(x).float())
                    text_parser = TextParser(tokenizer=tokenizer, post_proc_fn=lambda x: torch.from_numpy(x).int())

                    if isinstance(category, dict):
                        assert all([v in ["image", "text"] for v in category.values()])
                    else:
                        if category not in ["image", "text"]:
                            raise ValueError(f"Category: {category} is not supported yet. Raise a PR!")

                    if category == "text":
                        assert tokenizer != None, "tokenizer cannot be none for a text model!"

                    done = True
            except NameError:
                pass

            try:
                if isinstance(model_or_model_url, onnxruntime.InferenceSession):
                    logger.info(f"Trying to load from onnx model: {model_or_model_url}")
                    __framework = "onnx"

                    # we have to create templates using the nbox_meta
                    templates = None
                    if nbox_meta is not None:
                        all_inputs = nbox_meta["metadata"]["inputs"]
                        templates = {}
                        for node, meta in all_inputs.items():
                            templates[node] = [int(x["size"]) for x in meta["tensorShape"]["dim"]]

                    image_parser = ImageParser(post_proc_fn=lambda x: x.astype(np.float32), templates = templates)
                    text_parser = TextParser(tokenizer=tokenizer, post_proc_fn=lambda x: x.astype(np.int32))

                    self.session = model_or_model_url
                    self.input_names = [x.name for x in self.session.get_inputs()]
                    self.output_names = [x.name for x in self.session.get_outputs()]

                    logger.info(f"Inputs: {self.input_names}")
                    logger.info(f"Outputs: {self.output_names}")

                    done = True
            except NameError:
                pass

            if not done:
                raise ValueError(f"Unsupported model type: {type(model_or_model_url)}")


        # values coming from the blocks above
        self.model_or_model_url = model_or_model_url
        self.__framework = __framework
        self.text_parser = text_parser
        self.image_parser = image_parser
        self.nbox_meta = nbox_meta
        self.nbx_api_key = nbx_api_key
        self.category = category
        self.tokenizer = tokenizer
        self.model_key = model_key
        self.verbose = verbose
        self.__device = "cpu"
        self.cache_dir = gettempdir() if cache_dir == None else cache_dir

        logger.info(f"Model loaded successfully")
        logger.info(f"Model framework: {self.__framework}")
        logger.info(f"Model category: {self.category}")

    def fetch_meta_from_nbx_cloud(self):
        """When this is on NBX-Deploy cloud, fetch the metadata from the deployment node."""
        logger.info("Getting model metadata")
        URL = secret.get("nbx_url")
        r = requests.get(f"{URL}/api/model/get_model_meta", params=f"url={self.model_or_model_url}&key={self.nbx_api_key}")
        try:
            r.raise_for_status()
        except Exception as e:
            raise ValueError(f"Could not fetch metadata, please check status: {r.status_code}")

        # start getting the metadata, note that we have completely dropped using OVMS meta and instead use nbox_meta
        content = json.loads(r.content.decode())["meta"]
        nbox_meta = content["nbox_meta"]

        all_inputs = nbox_meta["metadata"]["inputs"]
        templates = {}
        for node, meta in all_inputs.items():
            templates[node] = [int(x["size"]) for x in meta["tensorShape"]["dim"]]

        if self.verbose:
            print("--------------")
            pp(nbox_meta)
            print("--------------")
            pp(templates)
            print("--------------")

        logger.info("Cloud infer metadata obtained")

        return nbox_meta, templates

    def __call__(self, input_object, return_inputs=False, return_dict=False, method=None, sklearn_args=None):
        r"""Caller is the most important UI/UX. The ``input_object`` can be anything from
        a tensor, an image file, filepath as string, string and must be processed automatically by a
        well written ``nbox.parser.BaseParser`` object . This ``__call__`` should understand the different
        usecases and manage accordingly.

        The current idea is that what ever the input, based on the category (image, text, audio, smell)
        it will be parsed through dedicated parsers that can make ingest anything.

        The entire purpose of this package is to make inference chill.

        Args:
            input_object (Any): input to be processed
            return_inputs (bool, optional): whether to return the inputs or not. Defaults to False.
            method(str, optional): specifically for sklearn models, this is the method to be called
                if nothing is provided then we call ``.predict()`` method.

        Returns:
            Any: currently this is output from the model, so if it is tensors and return dicts.
        """

        model_input = self._handle_input_object(input_object=input_object)

        if self.__framework == "nbx":
            logger.info(f"Hitting API: {self.model_or_model_url}")
            st = time()
            # OVMS has :predict endpoint and nbox has /predict
            _p = "/" if "export_type" in self.nbox_meta["spec"] else ":"
            json = {"inputs": model_input}
            if "export_type" in self.nbox_meta["spec"]:
                json["method"] = method
            r = requests.post(self.model_url + f"{_p}predict", json=json, headers={"NBX-KEY": self.nbx_api_key})
            et = time() - st
            out = None

            try:
                r.raise_for_status()
                out = r.json()

                # first try outputs is a key and we can just get the structure from the list
                if isinstance(out["outputs"], dict):
                    out = {k: np.array(v) for k, v in r.json()["outputs"].items()}
                elif isinstance(out["outputs"], list):
                    out = np.array(out["outputs"])
                else:
                    raise ValueError(f"Outputs must be a dict or list, got {type(out['outputs'])}")
                logger.info(f"Took {et:.3f} seconds!")
            except Exception as e:
                logger.info(f"Failed: {str(e)} | {r.content.decode()}")

        elif self.__framework == "sklearn" or self.__framework == "self-pkl":
            if "sklearn.neighbors.NearestNeighbors" in str(type(self.model_or_model_url)):
                method = getattr(self.model_or_model_url, "kneighbors") if method == None else getattr(self.model_or_model_url, method)
                out = method(model_input, **sklearn_args)
            elif "sklearn.cluster" in str(type(self.model_or_model_url)):
                if any(
                    x in str(type(self.model_or_model_url)) for x in ["AgglomerativeClustering", "DBSCAN", "OPTICS", "SpectralClustering"]
                ):
                    method = getattr(self.model_or_model_url, "fit_predict")
                    out = method(model_input)
            else:
                try:
                    method = getattr(self.model_or_model_url, "predict") if method == None else getattr(self.model_or_model_url, method)
                    out = method(model_input)
                except Exception as e:
                    print("[ERROR] Model Prediction Function is not yet registered " + e)

        elif self.__framework == "pytorch":
            with torch.no_grad():
                if isinstance(model_input, dict):
                    model_input = {k: v.to(self.__device) for k, v in model_input.items()}
                    out = self.model_or_model_url(**model_input)
                else:
                    assert isinstance(model_input, torch.Tensor)
                    model_input = model_input.to(self.__device)
                    out = self.model_or_model_url(model_input)

                # # bring back to cpu
                # if isinstance(out, (tuple, list)):
                #     out = [x.to("cpu") for x in out]
                # elif isinstance(out, torch.Tensor):
                #     out = out.to("cpu")
                # elif isinstance(out, dict):
                #     out = {k: v.to("cpu") for k, v in out.items()}

            # if self.nbox_meta is not None and self.nbox_meta.get("metadata", False) and self.nbox_meta["metadata"].get("outputs", False):
            #     outputs = self.nbox_meta["metadata"]["outputs"]
            #     if not isinstance(out, torch.Tensor):
            #         assert len(outputs) == len(out)
            #         out = {k: v.numpy() for k, v in zip(outputs, out)}
            #     else:
            #         out = {k: v.numpy() for k, v in zip(outputs, [out])}

        elif self.__framework == "onnx":
            if set(model_input.keys()) != set(self.input_names):
                diff = set(model_input.keys()) - set(self.input_names)
                return f"model_input keys do not match input_name: {diff}"
            out = self.model_or_model_url.run(self.output_names, model_input)
 
        # convert to dictionary if needed
        if return_dict:
            output_names = list(self.nbox_meta["metadata"]["outputs"].keys())
            if isinstance(out, (tuple, list)):
                out = {k: v for k, v in zip(output_names, out)}
            elif isinstance(out, dict):
                pass
            else:
                try:
                    if isinstance(out, (torch.Tensor, np.ndarray)):
                        out = {k: v.tolist() for k, v in zip(output_names, [out])}
                except NameError:
                    pass

                raise ValueError(f"Outputs must be a dict or list, got {type(out['outputs'])}")

        if return_inputs:
            return out, model_input
        return out

    def _handle_input_object(self, input_object):
        """First level handling to convert the input object to a fixed object"""
        # in case of scikit learn user must ensure that the input_object is model_input
        if self.__framework == "sklearn":
            return input_object

        elif self.__framework in ["nbx", "onnx"]:
            # the beauty is that the server is using the same code as this meaning that client
            # can get away with really simple API calls
            inputs_deploy = set(self.nbox_meta["metadata"]["inputs"].keys())
            if isinstance(input_object, dict):
                inputs_client = set(input_object.keys())
                assert inputs_deploy == inputs_client, f"Inputs mismatch, deploy: {inputs_deploy}, client: {inputs_client}"
                input_object = input_object
            else:
                if len(inputs_deploy) == 1:
                    input_object = {list(inputs_deploy)[0]: input_object}
                else:
                    assert len(input_object) == len(inputs_deploy), f"Inputs mismatch, deploy: {inputs_deploy}, client: {len(input_object)}"
                    input_object = {k: v for k, v in zip(inputs_deploy, input_object)}

        if isinstance(self.category, dict):
            assert isinstance(input_object, dict), "If category is a dict then input must be a dict"
            # check for same keys
            assert set(input_object.keys()) == set(self.category.keys())
            input_dict = {}
            for k, v in input_object.items():
                if k in self.category:
                    if self.category[k] == "image":
                        input_dict[k] = self.image_parser(v)
                    elif self.category[k] == "text":
                        input_dict[k] = self.text_parser(v)
                    else:
                        raise ValueError(f"Unsupported category: {self.category[k]}")
            return input_dict

        elif self.category == "image":
            input_obj = self.image_parser(input_object)
            return input_obj

        elif self.category == "text":
            # perform parsing for text and pass to the model
            input_dict = self.text_parser(input_object)
            return input_dict

        # Code below this part is super buggy and is useful for sklearn model,
        # please improve this as more usecases come up
        elif self.category == None:
            if isinstance(input_object, dict):
                return {k: v.tolist() for k, v in input_object.items()}
            return input_object.tolist()

        # when user gives a list as an input, it's better just to pass it as is
        # but when the input becomes a dict, this might fail.
        return input_object

    def get_nbox_meta(self, input_object, return_kwargs = True):
        """Get the nbox meta and trace args for the model with the given input object

        Args:
            input_object (Any): input to be processed
        """
        # this function gets the nbox metadata for the the current model, based on the input_object
        if self.__framework == "nbx":
            return self.nbox_meta

        args = None
        if self.__framework == "pytorch":
            args = inspect.getfullargspec(self.model_or_model_url.forward).args
            args.remove("self")

        self.eval()  # covert to eval mode
        model_output, model_input = self(input_object, return_inputs=True)

        # need to convert inputs and outputs to list / tuple
        dynamic_axes_dict = {
            0: "batch_size",
        }
        if self.category == "text":
            dynamic_axes_dict[1] = "sequence_length"

        # need to convert inputs and outputs to list / tuple
        if isinstance(model_input, dict):
            model_inputs = tuple(model_input.values())
            input_names = tuple(model_input.keys())
            input_shapes = tuple([tuple(v.shape) for k, v in model_input.items()])
        elif isinstance(model_input, (torch.Tensor, np.ndarray)):
            model_inputs = tuple([model_input])
            input_names = tuple(["input_0"]) if args is None else tuple(args)
            input_shapes = tuple([tuple(model_input.shape)])
        dynamic_axes = {i: dynamic_axes_dict for i in input_names}

        if isinstance(model_output, dict):
            output_names = tuple(model_output.keys())
            output_shapes = tuple([tuple(v.shape) for k, v in model_output.items()])
            model_output = tuple(model_output.values())
        elif isinstance(model_output, (list, tuple)):
            mo = model_output[0]
            if isinstance(mo, dict):
                # cases like [{"output_0": tensor, "output_1": tensor}]
                output_names = tuple(mo.keys())
                output_shapes = tuple([tuple(v.shape) for k, v in mo.items()])
            else:
                output_names = tuple([f"output_{i}" for i, x in enumerate(model_output)])
                output_shapes = tuple([tuple(v.shape) for v in model_output])
        elif isinstance(model_output, (torch.Tensor, np.ndarray)):
            output_names = tuple(["output_0"])
            output_shapes = (tuple(model_output.shape),)

        meta = get_meta(input_names, input_shapes, model_inputs, output_names, output_shapes, model_output)
        out = {
            "args": model_inputs,
            "outputs": model_output,
            "input_shapes": input_shapes,
            "output_shapes": output_shapes,
            "input_names": input_names,
            "output_names": output_names,
            "dynamic_axes": dynamic_axes,
        }
        if return_kwargs:
            return meta, out
        return meta

    def serialise(
        self,
        input_object,
        model_name=None,
        export_type="onnx",
        generate_ov_args = False,
        return_meta = False
    ):
        """This creates a singular .nbox file that contains the model binary and config file in ``self.cache_dir``

        Creates a folder at ``/tmp/{hash}`` and then adds three files to it:
            - ``/tmp/{hash}/{model_name}.{export_type}``
            - ``/tmp/{hash}/{model_name}.json``
            - ``/tmp/{hash}/{model_name}.nbox``

        and then later deletes the ``{model_name}.{export_type}`` and ``{model_name}.json`` and returns
        ``{model_name}.nbox`` and if required the metadata

        Args:
            input_object (Any): input to be processed
            model_name (str, optional): name of the model
            export_type (str, optional): [description]. Defaults to "onnx".
            generate_ov_args (bool, optional): Return the CLI args to be passed for Intel OpenVino
        """

        # First Step: check the args and see if conditionals are correct or not
        def __check_conditionals():
            assert self.__framework != "nbx", "This model is already deployed on the cloud"
            assert export_type in ["onnx", "torchscript", "pkl"], "Export type must be onnx, torchscript or pickle"
            if self.__framework == "sklearn":
                assert export_type in ["onnx", "pkl"], f"Export type must be onnx or pkl | got {export_type}"
            if self.__framework == "pytorch":
                assert export_type in ["onnx", "torchscript"], f"Export type must be onnx or torchscript | got {export_type}"

        # perform sanity checks on the input values
        __check_conditionals()

        # get metadata and exporing values
        nbox_meta, export_kwargs = self.get_nbox_meta(input_object, return_kwargs=True)
        _m_hash = utils.hash_(self.model_key)
        export_folder = utils.join(self.cache_dir, _m_hash)
        logger.info(f"Exporting model to {export_folder}")
        os.makedirs(export_folder, exist_ok=True)

        # intialise the console logger
        model_name = model_name if model_name is not None else f"{utils.get_random_name()}-{_m_hash[:4]}".replace("-", "_")
        export_model_path = utils.join(export_folder, f"{model_name}.{export_type}")
        logger.info("-" * 30 + f" Exporting {model_name}")

        # convert the model -> create a the spec, get the actual method for conversion
        logger.info(f"model_name: {model_name}")
        logger.info(f"Export type: {export_type}")
        
        try:
            export_fn = getattr(globals()[f"frm_{self.__framework}"], f"export_to_{export_type}", None)
            if export_fn == None:
                raise KeyError(f"Export type {export_type} not supported for {self.__framework}")
        except KeyError:
            raise ValueError(f"Framework {self.__framework} not supported (likely missing packages), check logs for more info")

        logger.info(f"Converting using: {export_fn}")
        export_fn(model=self.model_or_model_url, export_model_path=export_model_path, **export_kwargs)
        logger.info("Conversion Complete")

        # construct the output
        convert_args = None
        if generate_ov_args:
            # https://docs.openvinotoolkit.org/latest/openvino_docs_MO_DG_prepare_model_convert_model_Converting_Model_General.html
            input_ = ",".join(export_kwargs["input_names"])
            input_shape = ",".join([str(list(x.shape)).replace(" ", "") for x in export_kwargs["args"]])
            convert_args = f"--data_type=FP32 --input_shape={input_shape} --input={input_} "

            if self.category == "image":
                # mean and scale have to be defined for every single input
                # these values are calcaulted from uint8 -> [-1,1] -> ImageNet scaling -> uint8
                mean_values = ",".join([f"{name}[182,178,172]" for name in export_kwargs["input_names"]])
                scale_values = ",".join([f"{name}[28,27,27]" for name in export_kwargs["input_names"]])
                convert_args += f"--mean_values={mean_values} --scale_values={scale_values}"
            logger.info(convert_args)

        # write the nbox meta in a file and tarball the model and meta
        nbox_meta = {
            "metadata": nbox_meta,
            "spec": {
                "category": self.category,
                "model_key": self.model_key,
                "model_name": model_name,
                "src_framework": self.__framework,
                "export_type": export_type,
                "convert_args": convert_args
            },
        }
        meta_path = utils.join(export_folder, f"{model_name}.json")
        with open(meta_path, "w") as f:
            f.write(json.dumps(nbox_meta))

        nbx_path = os.path.join(export_folder, f"{model_name}.nbox")
        with tarfile.open(nbx_path, "w|gz") as tar:
            tar.add(export_model_path, arcname=os.path.basename(export_model_path))
            tar.add(meta_path, arcname=os.path.basename(meta_path))

        # remove the files
        os.remove(export_model_path)
        os.remove(meta_path)

        if return_meta:
            return nbx_path, nbox_meta
        return nbx_path

    @classmethod
    def deserialise(cls, filepath, verbose=False):
        """This is the counterpart of ``serialise``, it deserialises the model from the .nbox file.

        Args:
            nbx_path (str): path to the .nbox file
        """
        if not tarfile.is_tarfile(filepath) or not filepath.endswith(".nbox"):
            raise ValueError(f"{filepath} is not a valid .nbox file")

        with tarfile.open(filepath, "r:gz") as tar:
            folder = utils.join(gettempdir(), os.path.split(filepath)[-1].split(".")[0])
            tar.extractall(folder)

        files = glob(utils.join(folder, "*"))
        nbox_meta, nbox_meta_path, model_path = None, None, None
        for f in files:
            ext = f.split(".")[-1]
            if ext == "json":
                nbox_meta_path = f
                with open(f, "r") as f:
                    nbox_meta = json.load(f)
            elif ext in ["onnx", "pkl", "pt", "torchscript"]:
                model_path = f

        logger.info(f"nbox_meta: {nbox_meta_path}")
        logger.info(f"model_path: {model_path}")
        if not (nbox_meta or model_path):
            shutil.rmtree(folder)
            raise ValueError(f"{filepath} is not a valid .nbox file")

        export_type = nbox_meta["spec"]["export_type"]
        src_framework = nbox_meta["spec"]["src_framework"]
        category = nbox_meta["spec"]["category"]

        try:
            if export_type == "onnx":
                model = onnxruntime.InferenceSession(model_path)
            elif src_framework == "pt":
                if export_type == "torchscript":
                    model = torch.jit.load(model_path, map_location="cpu")
            elif src_framework == "sk":
                if export_type == "pkl":
                    with open(model_path, "rb") as f:
                        model = joblib.load(f)
        except Exception as e:
            raise ValueError(f"{export_type} not supported, are you missing packages? {e}")

        model = cls(model_or_model_url=model, category=category, nbox_meta=nbox_meta, verbose=verbose)
        shutil.rmtree(folder)
        return model

    def deploy(
        self,
        input_object,
        model_name=None,
        wait_for_deployment=False,
        runtime="onnx",
        deployment_type="nbox",
        deployment_id=None,
        deployment_name=None,
    ):
        """NBX-Deploy `read more <https://nimbleboxai.github.io/nbox/nbox.model.html>`_

        This deploys the current model onto our managed K8s clusters. This tight product service integration
        is very crucial for us and is the best way to make deploy a model for usage.

        Raises appropriate assertion errors for strict checking of inputs

        Args:
            input_object (Any): input to be processed
            model_name (str, optional): custom model name for this model
            cache_dir (str, optional): custom caching directory
            wait_for_deployment (bool, optional): wait for deployment to complete
            runtime (str, optional): runtime to use for deployment should be one of ``["onnx", "torchscript"]``, default is ``onnx``
            deployment_type (str, optional): deployment type should be one of ``['ovms2', 'nbox']``, default is ``nbox``
            deployment_id (str, optional): ``deployment_id`` to put this model under, if you do not pass this
                it will automatically create a new deployment check `platform <https://nimblebox.ai/oneclick>`_
                for more info or check the logs.
            deployment_name (str, optional): if ``deployment_id`` is not given and you want to create a new
                deployment group (ie. webserver will create a new ``deployment_id``) you can tell what name you
                want, be default it will create a random name.
        """
        # First Step: check the args and see if conditionals are correct or not
        def __check_conditionals():
            assert self.__framework != "nbx", "This model is already deployed on the cloud"
            assert deployment_type in ["ovms2", "nbox"], f"Only OpenVino and Nbox-Serving is supported got: {deployment_type}"
            if self.__framework == "sklearn":
                assert deployment_type == "nbox", "Only ONNX Runtime is supported for scikit-learn Framework"
            if deployment_type == "ovms2":
                assert runtime == "onnx", "Only ONNX Runtime is supported for OVMS2 Framework"
            if deployment_id and deployment_name:
                raise ValueError("deployment_id and deployment_name cannot be used together, pass only id")

        # perform sanity checks on the input values
        __check_conditionals()

        export_model_path, nbox_meta = self.serialise(
            input_object,
            model_name = model_name,
            export_type = runtime,
            generate_ov_args = deployment_type == "ovms2",
            return_meta = True
        )

        nbox_meta["spec"]["deployment_type"] = deployment_type
        nbox_meta["spec"]["deployment_id"] = deployment_id
        nbox_meta["spec"]["deployment_name"] = deployment_name

        # OCD baby!
        return deploy_model(
            export_model_path=export_model_path,
            nbox_meta=nbox_meta,
            wait_for_deployment=wait_for_deployment,
        )



from ..auth import AWSClient

def s3_download(client: AWSClient, bucket_name: str, key: str, local_path: str):
  s3c = client.get_client("s3")
  s3c.download_file(
    Bucket = bucket_name,
    Key = key,
    Filename = local_path
  )

__all__ = [
  "s3_download"
]


from .init import reset_log
reset_log()

from .model import Model
from .load import load, plug, PRETRAINED_MODELS
from .parsers import ImageParser, TextParser
from .jobs import Instance
from .auth import AWSClient, GCPClient, OCIClient, DOClient, AzureClient
from .operators import Operator

__version__ = "0.8.4"



# MIT-License code for all operators that are open sourced

from ast import arg
from .operator import Operator

from ..utils import Pool
from ..jobs import Instance

# nbox/

class NboxInstanceStartOperator(Operator):
  def __init__(self, instances):
    super().__init__()
    if not isinstance(instances, list):
      instances = [instances]
    assert instances[0].__class__ == Instance, "instances must be of type Instance"
    self.instances = instances
    self.pool = Pool("thread", len(instances), _name = "instance_starter")

  def forward(self):
    self.pool(
      lambda instance: instance.start(cpu_only = True),
      self.instances
    )
    return None


class NboxModelDeployOperator(Operator):
  def __init__(self, model_name, model_path, model_weights, model_labels):
    super().__init__()
    self.model_name = model_name
    self.model_path = model_path
    self.model_weights = model_weights
    self.model_labels = model_labels

  def forward(self, name):
    from ..model import Model
    Model(
      self.model_name,
      self.model_path,
      self.model_weights,
      self.model_labels,
    ).deploy(name)


class NboxWaitTillJIDComplete(Operator):
  def __init__(self, instance, jid):
    super().__init__()
    self.instance = instance
    self.jid = jid

  def forward(self, poll_interval = 5):
    status = self.instance(self.jid)
    if status == "done":
      return None
    elif status == "error-done":
      raise Exception("Job {} failed".format(self.jid))
    elif status == "running":
      from time import sleep
      while status == "running":
        sleep(poll_interval)
        status = self.instance(self.jid)
      if status == "done":
        return None
      elif status == "error-done":
        raise Exception("Job {} failed".format(self.jid))

# /nbox


# arch/

class Sequential():
  def __init__(self, *ops):
    super().__init__()
    for op in ops:
      assert isinstance(op, Operator), "Operator must be of type Operator"
    self.ops = ops

  def forward(self, x = None, capture_output = False):
    out = x
    outputs = []
    for op in self.ops:
      out = op(out)
      outputs.append(out)
    if capture_output:
      return out, outputs
    return out

# /arch

class Python(Operator):
  def __init__(self, func, *args, **kwargs):
    super().__init__()
    self.fak = (func, args, kwargs)

  def forward(self):
    return self.fak[0](*self.fak[1], **self.fak[2])

class GitClone(Operator):
  def __init__(self, url, path = None, branch = None):
    super().__init__()
    self.url = url
    self.path = path
    self.branch = branch

  def forward(self):
    # git clone -b <branch> <repo> <path>
    import subprocess
    command = ["git", "clone"]
    if self.branch:
      command.append("-b")
      command.append(self.branch)
    if self.path:
      command.append(self.path)
    command.append(self.url)
    subprocess.run(command, check = True)

class ShellCommand(Operator):
  def __init__(self, *commands):
    super().__init__()
    import string

    self.commands = commands
    all_in = []
    for c in self.commands:
      all_in.extend([tup[1] for tup in string.Formatter().parse(c) if tup[1] is not None])
    self._inputs = all_in

  def forward(self, *args, **kwargs):
    import shlex
    import subprocess
    for comm in self.commands:
      comm = comm.format(*args, **kwargs)
      comm = shlex.split(comm)
      subprocess.run(comm, check = True)

class Notify(Operator):
  _mode_to_packages = {
    "slack": "slackclient",
    "ms_teams": "microsoft-teams",
    "discord": "discord",
  }

  def __init__(
    self,
    slack_connect: str = None,
    ms_teams: str = None,
    discord: str = None,
  ):
    super().__init__()
    self.notify_mode = None
    self.notify_id = None

    for mode, id in [
      ("slack", slack_connect),
      ("ms_teams", ms_teams),
      ("discord", discord),
    ]:
      if id:
        self.notify_mode = mode
        self.notify_id = id

        # check for package dependencies
        import importlib
        try:
          importlib.import_module(self._mode_to_packages[mode])
        except ImportError:
          raise Exception(f"{self._mode_to_packages[mode]} package required for {mode}")
        break
    
    if not self.notify_mode:
      raise Exception("No notification mode specified")

  def forward(self, message: str, **kwargs):
    package = self._mode_to_packages[self.notify_mode]
    import importlib
    importlib.import_module(package)
    if self.notify_mode == "slack":
      from slackclient import SlackClient
      sc = SlackClient(self.notify_id)
      sc.api_call("chat.postMessage", text = message, **kwargs)
    elif self.notify_mode == "ms_teams":
      from microsoft_teams.api_client import TeamsApiClient
      from microsoft_teams.models import MessageCard
      client = TeamsApiClient(self.notify_id)
      client.connect()
      client.send_message(MessageCard(text = message, **kwargs))
    elif self.notify_mode == "discord":
      from discord import Webhook, RequestsWebhookAdapter
      webhook = Webhook.from_url(self.notify_id, adapter = RequestsWebhookAdapter())
      webhook.send(message, **kwargs)



import logging

from .jobs import (
  get_status, Instance, Subway
)
from ..utils import nbox_session

logger = logging.getLogger()
web_server_subway = Subway("https://nimblebox.ai", nbox_session)

def init(project_name):
  print("-" * 69)
  import os, re
  from datetime import datetime
  created_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

  out = re.findall("^[a-zA-Z_]+$", project_name)
  if not out:
    raise ValueError("Project name can only contain letters and underscore")

  if os.path.exists(project_name):
    raise ValueError(f"Project {project_name} already exists")
  
  print(f"Creating a folder: {project_name}")
  os.mkdir(project_name)
  os.chdir(project_name)

  print("Creating: exe.py")
  with open("exe.py", "w") as f:
    code_ = f'''
#!/usr/bin/env python3
# auto generated by 'nbox jobs init' command
# project_name: {project_name}
# created_time: {created_time}

from nbox import Operator
from nbox.utils import folder

def get_op() -> Operator:
  # since initialising your operator might require passing a bunch of arguments
  # you can use this function to get the operator by manually defining things here

  return None

def deploy():
  # get operator and deploy it
  op: Operator = get_op()
  op.deploy(
    job_id = "if job_id is None, new job will be created",
    init_folder = folder(__file__) # no need to change this
  )

def run():
  # run your operator
  op: Operator = get_op()
  op(
    # if your operator is scheduled, avoid passing anything here
    # if your operator is run one time, pass the input here
  )
'''

    code_ += '''
if __name__ == "__main__":
  from fire import Fire
  Fire({"deploy": deploy, "run": run})

# end of auto generated code'''
    f.write(code_.strip())

  with open("README.md", "w") as f:
    f.write(f'''
# '{project_name}' NBX-Jobs

This is an autogenerated README file for [NBX-Jobs](www.nimblebox.ai/jobs) which was:
* created at: {created_time}
* name: {project_name}

Steps to complete:

1. Add all your code module here in this folder
2. Add in requirements.txt file if needed
3. Fill The Three functions (`get_operator`, `deploy`, `run`) in exe.py
4. Run `nbox jobs deploy <folder_path> or `nbox jobs deploy -h` for more help

Rule of Thumb:

1. If you are scheduling, your `op` forward should not takes arguments, otherwise
you are executing the same process over and over, which is not what you want.
If you still want to do so, try moving as much as possible to `__init__`.
2. If you are running this one-time, then you should define arguments in `run` function.

    '''.strip() + "\n")

  open("requirements.txt", "w").close()

  print("Completed")
  print("-" * 69)

def deploy(folder: str = "./"):
  """Deploy a job on NimbleBox.ai's NBX-Jobs. This convinience function will
  just run ``./exe.py deploy``

  Args:
    folder (str, optional): Folder to deploy. Defaults to "./".
  """

  import os
  import sys
  import subprocess
  
  if not os.path.exists(folder) or not os.path.isdir(folder):
    raise ValueError(f"Incorrect project at path: '{folder}'! nbox jobs init <name>")
  if os.path.isdir(folder):
    os.chdir(folder)
    if not os.path.exists("./exe.py"):
      raise ValueError(f"Incorrect project at path: '{folder}'! nbox jobs init <name>")
  else:
    raise ValueError(f"Incorrect project at path: '{folder}'! nbox jobs init <name>")

  sys.path.append(folder)
  os.chdir(folder)
  subprocess.call(["./exe.py", "deploy"])

def status(id_or_name: str):
  raise NotImplementedError

def new(name: str):
  Instance.create(name)


# -------------------------




# singluar loader file for all models in nbox

import re
import json
import inspect
from typing import Dict

from .model import Model
from .utils import _isthere, fetch

import logging

logger = logging.getLogger()

model_key_regex = re.compile(r"^(\w+)(\/[\w\/-]+)?(:*[\w+:]+)?$")

# util functions
def remove_kwargs(model_fn, kwargs):
    """take in the ``model_fn`` and only keep variables from kwargs that can be consumed"""
    # compare variables between the model_fn and kwargs if they are different then remove it with warning
    arg_spec = inspect.getfullargspec(model_fn)
    if kwargs and arg_spec.varkw != None:
        diff = set(kwargs.keys()) - set(arg_spec.args)
        for d in list(diff):
            kwargs.pop(d)
    return kwargs


# --- model loader functions: add your things here
# guide: all models are indexed as follows
# {
#   "key": (builder_function, "category")
#
#   # to be moved to
#   "key": (builder_function, "task_type", "source", "pre", "task", "post")
# }
#
# Structure of each loader function looks as follows:
# def loader_fn() -> <dict as above>
#
# Each model builder function looks as follows:
# def model_builder() -> (model, model_kwargs)


def load_efficientnet_pytorch_models() -> Dict:
    def model_builder(pretrained=False, **kwargs):
        import efficientnet_pytorch

        if pretrained:
            model_fn = efficientnet_pytorch.EfficientNet.from_pretrained
        else:
            model_fn = efficientnet_pytorch.EfficientNet.from_name

        kwargs = remove_kwargs(model_fn, kwargs)
        return model_fn(**kwargs), {}

    return {"efficientnet_pytorch/efficientnet": (model_builder, "image")}


def load_torchvision_models() -> Dict:
    def model_builder(model, pretrained=False, **kwargs):
        # tv_mr = torchvision models registry
        # this is a json object that maps all the models to their respective methods from torchvision
        # the trick to loading is to eval the method string
        tv_mr = json.loads(fetch("https://raw.githubusercontent.com/NimbleBoxAI/nbox/master/assets/pt_models.json").decode("utf-8"))
        model_fn = tv_mr.get(model, None)
        if model_fn == None:
            raise IndexError(f"Model: {model} not found in torchvision")

        # torchvision is imported here, if it is imported in the outer method, eval fails
        # reason: unknown
        import torchvision

        model_fn = eval(model_fn)
        kwargs = remove_kwargs(model_fn, kwargs)
        model = model_fn(pretrained=pretrained, **kwargs)
        return model, {}

    return {"torchvision": (model_builder, "image")}


def load_transformers_models() -> Dict:
    def hf_model_builder(model, model_instr, **kwargs):
        import transformers

        _auto_loaders = {x: getattr(transformers, x) for x in dir(transformers) if x[:4] == "Auto" and x != "AutoConfig"}

        model_instr = model_instr.split("::")
        if len(model_instr) == 1:
            auto_model_type = model_instr[0]
        else:
            # if the task is given, validate that as well
            auto_model_type, task = model_instr
            assert task in [
                "generation",
                "masked_lm",
            ], "For now only the following are supported: `generation`, `masked_lm`"

        # initliase the model and tokenizer object
        tokenizer = transformers.AutoTokenizer.from_pretrained(model, **kwargs)

        # All the tokenizers must contain the pad_token, by default we set it to eos_token
        if not tokenizer.pad_token:
            tokenizer.add_special_tokens({"pad_token": tokenizer.eos_token})

        # tokenizer.pad_token = tokenizer.eos_token if not tokenizer.pad_token_id else tokenizer.pad_token
        model = _auto_loaders[auto_model_type].from_pretrained(model, **kwargs)

        # return the model and tokenizer
        return model, {"tokenizer": tokenizer}

    return {"transformers": (hf_model_builder, "text")}


### ----- pretrained models master index
# add code based on conditionals, best way is to only include those that
# have proper model building code like transformers, torchvision, etc.

PRETRAINED_MODELS = {}
all_repos = ["efficientnet_pytorch", "torchvision", "transformers"]

logger.info(f"Loading plugins")
for repo in all_repos:
    if _isthere(repo):
        PRETRAINED_MODELS.update(locals()[f"load_{repo}_models"]())

# if there are no pretrained models available, then raise an error
if not PRETRAINED_MODELS:
    logging.warning("No pretrained models available")

# ---- plug your own models and in extension the methods here


def plug(src_name, builder_fn, cataegory):
    """Plug your nbox_builder methods here. Once plugged in your codebase, you can use the
    simplicity of nbox.loaders

    Args:
        src_name (str): name of the source
        builder_fn (func): function to be called to build the model
        cataegory (dict): input categories for the input
d
    Raises:
        ValueError: if src_name is already in the index
    """
    # check if the source is already present
    if src_name in PRETRAINED_MODELS:
        logging.warning(f"Source: {src_name} already present in the pretrained models index")
        return

    PRETRAINED_MODELS[src_name] = (builder_fn, cataegory) # add the source


# ---- load function has to manage everything and return Model object properly initialised


def load(model_key_or_url, verbose=False, **loader_kwargs):
    """This function loads the nbox.Model object from the pretrained models index.

    Args:

        registry (str):  key for which to load the model, the structure looks as follows:

            .. code-block:: python

                "source/(source/key)::<pre::task::post>"


        nbx_api_key (str, optional): If model_key_or_url has type `url` then pass the key corresponding url
        verbose (bool, optional): If True, prints logs
        cloud_infer (bool, optional): If true uses Nimblebox deployed inference and logs in
            using `nbx_api_key`. Defaults to False.
        loader_kwargs (dict, optional): keyword arguments to be passed to the loader function.

    Raises:
        ValueError: If `source` is not found
        IndexError: If `source` is found but `source/key` is not found

    Returns:
        nbox.Model: when using local inference
        nbox.NBXApi: when using cloud inference
    """
    # the input key can also contain instructions on how to run a particular models and so
    model_key_parts = re.findall(model_key_regex, model_key_or_url)
    if not model_key_parts:
        raise ValueError(f"Key: {model_key_or_url} incorrect, please check!")

    # this key is valid, now get it's components
    src, src_key, model_instr = model_key_parts[0]
    src_key = src_key.strip("/")  # remove leading and trailing slashes
    model_instr = model_instr.replace(":", "")  # remove the :
    model_key = model_key_or_url

    # print("  model_key:", model_key)
    # print("        src:", src)
    # print("    src_key:", src_key)
    # print("model_instr:", model_instr)

    model_fn, model_meta = PRETRAINED_MODELS.get(src, (None, None))
    if model_meta is None:
        model_fn, model_meta = PRETRAINED_MODELS.get(src, (None, None))
        if model_meta is None:
            raise IndexError(f"Model: {src} not found")

    # now just load the underlying graph and the model and off you go
    model, model_kwargs = model_fn(model=src_key, model_instr=model_instr, **loader_kwargs)
    out = Model(model_or_model_url=model, category=model_meta, model_key=model_key, verbose=verbose, **model_kwargs)

    return out



# this file has the utilities and functions required for processing Operator items
# to Airlfow specific operations

import inspect # used for __doc__
from functools import partial
from datetime import timedelta
from nbox.utils import isthere


def to_airflow_operator(operator, timeout: timedelta = None, operator_kwargs = {}):
  """
  Args:
      operator (Operator): nbox.Operator object to be converted to Airflow Operator
      timeout (timedelta, default=None): in how many seconds the operator should timeout
  """
  operator_kwargs_dict = {}

  # we consider our SLA as the timeout
  # ----------------------------------
  # :param execution_timeout: max time allowed for the execution of
  #      this task instance, if it goes beyond it will raise and fail.
  # :param sla: time by which the job is expected to succeed. Note that
  #      this represents the ``timedelta`` after the period is closed. For
  #      example if you set an SLA of 1 hour, the scheduler would send an email
  #      soon after 1:00AM on the ``2016-01-02`` if the ``2016-01-01`` instance
  #      has not succeeded yet.
  #      The scheduler pays special attention for jobs with an SLA and
  #      sends alert
  #      emails for SLA misses. SLA misses are also recorded in the database
  #      for future reference. All tasks that share the same SLA time
  #      get bundled in a single email, sent soon after that time. SLA
  #      notification are sent once and only once for each task instance.
  operator_kwargs_dict["execution_timeout"] = timeout
  operator_kwargs_dict["sla"] = timeout
  operator_kwargs_dict["task_id"] = operator.__class__.__name__

  try:
    comms = operator.comms()
  except:
    comms = {}

  # this is currently assumed to be rst text because this is what we are using
  # for documentation at NBX
  doc = operator.__doc__ # this is in between class xxx and def __init__
  doc = doc if doc else ""
  init_doc = inspect.getdoc(operator.__init__) # this is doc for __init__
  init_doc = init_doc if init_doc else ""
  full_doc = doc + "\n" + init_doc
  full_doc = None if not full_doc else full_doc
  operator_kwargs_dict["doc_rst"] = full_doc

  # update the dict with information
  operator_kwargs_dict.update(comms)
  operator_kwargs_dict.update(operator_kwargs)

  # importing inside the function  
  from airflow.models.baseoperator import BaseOperator
  from ..init import reset_log
  reset_log()

  operator = BaseOperator(
    # items planned to be supported
    email = None,
    email_on_retry = False,
    email_on_failure = False,
    
    # documentation to be handled, currently assumed to be rst documentation
    doc = None,
    doc_md = None,
    doc_json = None,
    doc_yaml = None,

    # these can be the hooks that user defines for Operators
    # ie. is the airflow.operator executes user defines which of
    # their functions to call
    on_execute_callback = None,
    on_failure_callback = None,
    on_success_callback = None,
    on_retry_callback = None,

    # others are ignored as of now
    **operator_kwargs_dict
  )

  return operator

def from_airflow_operator(operator_cls, air_operator):
  assert air_operator.__class__.__name__ != "PythonOperator", \
    "Only PythonOperator is supported at the moment"
  op = operator_cls()
  op._register_forward(air_operator.python_callable)
  return op

def to_airflow_dag(operator, dag, operator_kwrags, dag_kwargs):
  from airflow.models.dag import DAG
  from ..init import reset_log
  reset_log()

  operator = to_airflow_operator(operator, **operator_kwrags)
  dag = DAG(dag_id = "DAG_" + operator.task_id, **dag_kwargs)
  return dag

def from_airflow_dag(operator_cls, dag):
  task_group = dag.task_group # the DAG structure is inside dag.task_group

  topo_tree = {}
  for x in task_group:
    if x.__class__.__name__ != "PythonOperator":
      raise Exception("Only PythonOperator is supported at the moment")
    topo_tree[x.task_id] = [_x.task_id for _x in x.get_direct_relatives(upstream = True)]

  # from this topo tree create nbox operator for this dag
  root = operator_cls()
  all_ops = {}
  parent_less_ops = []
  for child, parents in topo_tree.items():
    if child not in all_ops:
      all_ops[child] = operator_cls()
      all_ops[child]._register_forward(dag.task_dict[child].python_callable)
    for p in parents:
      if p not in all_ops:
        all_ops[p] = operator_cls()
      setattr(
        all_ops[p], f"op__{p}__{child}", all_ops[child]
      )
      try:
        all_ops[p]._register_forward(dag.task_dict[p].python_callable)
      except:
        pass

    if not parents:
      parent_less_ops.append(child)

  for p in parent_less_ops:
    name = f"op__root__{p}"
    setattr(root, name, all_ops[p])

  def sequential_forward(ops, **kwargs):
    out = kwargs
    for o in ops:
      out = o(**out)
      if not out:
        out = {}
    return out

  forward = partial(sequential_forward, ops = root.operators())
  root._register_forward(forward)

  return root


class AirflowMixin:
  @isthere("airflow", soft = False)
  def to_airflow_operator(self, timeout, **operator_kwargs):
    return to_airflow_operator(self, timeout, operator_kwargs)

  @classmethod
  @isthere("airflow", soft = False)
  def from_airflow_operator(cls, air_operator):
    return from_airflow_operator(cls, air_operator)

  @isthere("airflow", soft = False)
  def to_airflow_dag(self, dag_kwargs, operator_kwargs):
    return to_airflow_dag(self, dag_kwargs, operator_kwargs)

  @classmethod
  @isthere("airflow", soft = False)
  def from_airflow_dag(cls, dag):
    return from_airflow_dag(cls, dag)


__all__ = ["AirflowMixin"]



"""
nboxCLI
=======

This is CLI for ``nbox``, it is meant to be as simple to use as possible.
The commands are broken down according to the products they are related to.

.. code-block::

  nbox deploy [OPTIONS]
  nbox instance [OPTIONS-1] [create/start/stop] [OPTIONS-2]
  nbox jobs [init/]
"""

import fire
from . import cli as n # nbox-cli
from .jobs import cli as j # jobx-cli
from .jobs import Instance

if __name__ == "__main__":
  fire.Fire({
    "deploy": n.deploy,           # nbox deploy
    "instance": Instance,         # nbox jobs instance
    "jobs": {
      "new": j.init,              # nbox jobs init PROJECT_NAME
      "deploy": j.deploy,         # nbox jobs deploy FOLDER
      "status": j.status,         # nbox jobs status ID_OR_NAME
    },
    "status": n.status,           # nbox status
    "tunnel": n.tunnel,
  })



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nbox', 'nbox.auto', 'nbox.framework', 'nbox.jobs', 'nbox.operators']

package_data = \
{'': ['*']}

install_requires = \
['randomname>=0.1.3,<0.2.0', 'requests>=2.25.1,<3.0.0', 'tabulate==0.8.9']

setup_kwargs = {
    'name': 'nbox',
    'version': '0.8.4',
    'description': 'ML Inference 🥶',
    'long_description': None,
    'author': 'NBX Research',
    'author_email': 'research@nimblebox.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# weird file name, yessir. Why? because
# from nbox.framework.on_functions import PureFunctionParser
# read as "from nbox's framework on Functions import the pure-function Parser"

import ast
import inspect
from uuid import uuid4
from logging import getLogger
logger = getLogger()

# ==================

# dataclasses are not that good: these classes are for the Op

class DBase:
  def __init__(self, **kwargs):
    for k, v in kwargs.items():
      setattr(self, k, v)
  
  def get_dict(self):
    data = {}
    for k in self.__slots__:
      _obj = getattr(self, k)
      if isinstance(_obj, DBase):
        data[k] = _obj.get_dict()
      elif _obj and isinstance(_obj, list) and isinstance(_obj[0], DBase):
        data[k] = [_obj.get_dict() for _obj in _obj]
      else:
        data[k] = _obj
    return data

class ExpressionNodeInfo(DBase):
  __slots__ = [
    'name', # :str
    'code', # :str (base64)
    'nbox_string', # :str
    'lineno', # :int
    'col_offset', # :int
    'end_lineno', # :int
    'end_col_offset', # :int
    'inputs', # :list[Dict[Any, Any]]
    'outputs', # :list[str]
  ]

class IfNodeInfo(DBase):
  __slots__ = [
    'nbox_string', # :str
    'conditions', # :list[ExpressionNodeInfo]
    'inputs', # :list[Dict[Any, Any]]
    'outputs', # :list[str]
  ]


# these classes are for the FE
class RunStatus(DBase):
  __slots__ = [
    'start', # :str
    'end', # :str
    'inputs', # :Dict
    'outputs', # :Dict
  ]


class Node(DBase):
  __slots__ = [
    'id', # :str
    'execution_index', # :int
    'name', # :str
    'type', # :str
    'node_info', # :Union[ExpressionNodeInfo, IfNodeInfo]
    'operator', # :str
    'nbox_string', # :str
    'run_status', # :RunStatus
  ]

class Edge(DBase):
  __slots__ = [
    'id', # :str
    'source', # :str
    'target', # :str
    'type', # :str
    'nbox_string', # :str
  ]


# ==================
# the next set of functions are meant as support methods to create nbox_strings IR.

class NboxStrings:
  OP_TO_STRING = {
    "function": "FUNCTION: {name} ( {inputs} ) => [ {outputs} ]",
    "define": "DEFINE: {name} ( {inputs} )",
    "for": "FOR: {name} ( {iter} ) => ( {target} )",
  }

  def __init__(self):
    pass

  def function(self, name, inputs, outputs):
    return self.OP_TO_STRING["function"].format(
      name=name,
      inputs=", ".join([f"{x['kwarg']}={x['value']}" for x in inputs]),
      outputs=", ".join(outputs)
    )

  def define(self, name, inputs):
    return self.OP_TO_STRING["define"].format(
      name=name,
      inputs=", ".join([f"{x['kwarg']}={x['value']}" for x in inputs])
    )

  def _for(self, name, iter, target):
    return self.OP_TO_STRING["for"].format(
      name=name,
      iter=iter,
      target=target
    )

nbxl = NboxStrings()

def write_program(nodes):
  for i, n in enumerate(nodes):
    if n.nbox_string == None:
      print(f"{i:03d}|{n.node_info.nbox_string}")
    else:
      print(f"{i:03d}|{n.nbox_string}")

# ==================

def get_code_portion(cl, lineno, col_offset, end_lineno, end_col_offset, b64 = True, **_):
  sl, so, el, eo = lineno, col_offset, end_lineno, end_col_offset
  if sl == el:
    return cl[sl-1][so:eo]
  code = ""
  for i in range(sl - 1, el, 1):
    if i == sl - 1:
      code += cl[i][so:]
    elif i == el - 1:
      code += "\n" + cl[i][:eo]
    else:
      code += "\n" + cl[i]
  
  # convert to base64
  if b64:
    import base64
    return base64.b64encode(code.encode()).decode()
  return code

def parse_args(node):
  inputs = []
  for a in node.args:
    a = a.arg if isinstance(a, ast.arg) else a
    inputs.append({
      "kwarg": None,
      "value": a,
    })
  for a in node.kwonlyargs:
    inputs.append({
      "kwarg": a[0],
      "value": a[1],
    })
  if node.vararg:
    inputs.append({
      "kwarg": "*"+node.vararg.arg,
      "value": None
    })
  if node.kwarg:
    inputs.append({
      "kwarg": "**"+node.kwarg.arg,
      "value": None
    })
  return inputs

def get_name(node):
  if isinstance(node, ast.Name):
    return node.id
  elif isinstance(node, ast.Attribute):
    return get_name(node.value) + "." + node.attr
  elif isinstance(node, ast.Call):
    return get_name(node.func)

def parse_kwargs(node, lines):
  if isinstance(node, ast.Name):
    return node.id
  if isinstance(node, ast.Constant):
    val = node.value
    return val
  if isinstance(node, ast.keyword):
    arg = node.arg
    value = node.value
    if 'id' in value.__dict__:
      # arg = my_model
      return (arg, value.id)
    elif 'value' in value.__dict__:
      if isinstance(value.value, ast.Call):
        # arg = my_model(...)
        return (arg, get_code_portion(lines, b64 = False, **value.value.__dict__))
      else:
        # arg = 20
        return (arg, value.value)
    elif 'func' in value.__dict__:
      #   arg = some_function(with, some=args)
      #   ^^^   ^^^^^
      # kwarg   value
      return {"kwarg": arg, "value": get_code_portion(lines, b64 = False, **value.__dict__)}
  if isinstance(node, ast.Call):
    return get_code_portion(lines, **node.func.__dict__)

def node_assign_or_expr(node, lines):
  # print(get_code_portion(lines, **node.__dict__))
  value = node.value
  try:
    name = get_name(value.func)
  except AttributeError:
    return None
  args = [parse_kwargs(x, lines) for x in value.args + value.keywords]
  inputs = []
  for a in args:
    if isinstance(a, dict):
      inputs.append(a)
      continue
    inputs.append({
      "kwarg": a[0] if isinstance(a, tuple) else None,
      "value": a[1] if isinstance(a, tuple) else a,
    })

  outputs = []
  if isinstance(node, ast.Assign):
    targets = node.targets[0]
    outputs = [parse_kwargs(x, lines) for x in targets.elts] \
      if isinstance(targets, ast.Tuple) \
      else [parse_kwargs(targets, lines)
    ]

  return ExpressionNodeInfo(
    name = name,
    inputs = inputs,
    outputs = outputs,
    nbox_string = nbxl.function(name, inputs, outputs),
    code = get_code_portion(lines, **node.__dict__),
    lineno = node.lineno,
    col_offset = node.col_offset,
    end_lineno = node.end_lineno,
    end_col_offset = node.end_col_offset,
  )

def node_if_expr(node, lines):
  def if_cond(node, lines, conds = []):
    if not hasattr(node, "test"):
      else_cond = list(filter(lambda x: x["condition"] == "else", conds))
      if not else_cond:
        conds.append({
          "condition": "else",
          "code": dict(
            lineno = node.lineno,
            col_offset = node.col_offset,
            end_lineno = node.end_lineno,
            end_col_offset = node.end_col_offset,
          )
        })
      else:
        cond = else_cond[0]
        cond["code"]["end_lineno"] = node.end_lineno
        cond["code"]["end_col_offset"] = node.end_col_offset
    else:
      condition = get_code_portion(lines, **node.test.__dict__)

      # need to run this last or "else" comes up first
      conds.append({
        "condition": condition,
        "code": {
          "lineno": node.lineno,
          "col_offset": node.col_offset,
          "end_lineno": node.end_lineno,
          "end_col_offset": node.end_col_offset,
        }
      })
      for x in node.orelse:
        if_cond(x, lines, conds)

    return conds
  
  # get all the conditions and structure as ExpressionNodeInfo
  
  all_conditions = if_cond(node, lines, conds = [])
  ends = []
  for b0, b1  in zip(all_conditions[:-1], all_conditions[1:]):
    ends.append([b0["code"], b1["code"]])
  for i in range(len(ends)):
    ends[i] = {
      "lineno": ends[i][0]["lineno"],
      "col_offset": ends[i][0]["col_offset"],
      "end_lineno": ends[i][1]["lineno"],
      "end_col_offset": ends[i][1]["col_offset"],
    }
  ends += [all_conditions[-1]["code"]]

  conditions = []
  for i, c in enumerate(all_conditions):
    box = ends[i]
    _node = ExpressionNodeInfo(
      name = f"if-{i}",
      nbox_string = c["condition"],
      code = get_code_portion(lines, **box),
      lineno = box['lineno'],
      col_offset = box['col_offset'],
      end_lineno = box['end_lineno'],
      end_col_offset = box['end_col_offset'],
      inputs = [],
      outputs = [],
    )
    conditions.append(_node)

  nbox_string = "IF: { " + ", ".join(x.nbox_string for x in conditions) + " }"
  return IfNodeInfo(
    conditions = conditions,
    nbox_string = nbox_string,
    inputs = [],
    outputs = []
  )

def def_func_or_class(node, lines):
  out = {"name": node.name, "code": get_code_portion(lines, **node.__dict__), "type": "def-node"}
  if isinstance(node, ast.FunctionDef):
    out.update({"func": True, "inputs": parse_args(node.args)})
  else:
    out.update({"func": False, "inputs": []})
  return out


# ==================

type_wise_logic = {
  # defns ------
  ast.FunctionDef: def_func_or_class,
  ast.ClassDef: def_func_or_class,

  # nodes ------
  ast.Assign: node_assign_or_expr,
  ast.Expr: node_assign_or_expr,
  ast.If: node_if_expr,

  # todos ------
  # ast.AsyncFunctionDef: async_func_def,
  # ast.Await: node_assign_or_expr,
}

# ==================

def get_nbx_flow(forward):
  # get code string from operator
  code = inspect.getsource(forward).strip()
  code_lines = code.splitlines()
  node = ast.parse(code)

  edges = [] # this is the flow
  nodes = [] # this is the operators
  symbols_to_nodes = {} # this is things that are defined at runtime

  for i, expr in enumerate(node.body[0].body):
    # if isinstance(expr, ast.Module):
    #   continue

    if not type(expr) in type_wise_logic:
      # code pieces that are not yet supported should still see the code
      node = Node(
        id = str(uuid4()),
        execution_index = i,
        name = f"codeblock-{i}",
        type = "op-node",
        operator = "CodeBlock",
        node_info = ExpressionNodeInfo(
          name = f"codeblock-{i}", # :str
          code = get_code_portion(code_lines, bs64 = True, **expr.__dict__), # :str (base64)
          nbox_string = None, # :str
          lineno = expr.lineno, # :int
          col_offset = expr.col_offset, # :int
          end_lineno = expr.end_lineno, # :int
          end_col_offset = expr.end_col_offset, # :int
          inputs = [], # :list[Dict[Any, Any]]
          outputs = [], # :list[str]
        ),
      nbox_string = f"CODE: {str(type(expr))}", # :str
      run_status = RunStatus(start = None, end = None, inputs = [], outputs = [])
      )
      nodes.append(node)
      continue

    output = type_wise_logic[type(expr)](expr, code_lines)
    if output is None:
      continue

    if isinstance(output, ExpressionNodeInfo):
      output = Node(
        id = str(uuid4()),
        execution_index = i,
        name = output.name,
        type = "op-node",
        operator = "CodeBlock",
        node_info = output,
        nbox_string = None,
        run_status = RunStatus(start = None, end = None, inputs = [], outputs = [])
      )
      nodes.append(output)
    elif isinstance(output, IfNodeInfo):
      output = Node(
        id = str(uuid4()),
        execution_index = i,
        name = f"if-{i}",
        type = "op-node",
        operator = "Conditional",
        node_info = output,
        nbox_string = output.nbox_string,
        run_status = RunStatus(start = None, end = None, inputs = [], outputs = [])
      )
      nodes.append(output)
    elif "def" in output["type"]:
      symbols_to_nodes[output['name']] = {
        "node_info": output,
        "execution_index": i,
        "nbox_string": nbxl.define(output["name"], output["inputs"])
      }

  # edges for execution order can be added
  for op0, op1 in zip(nodes[:-1], nodes[1:]):
    edges.append(
      Edge(
        id = f"edge-{op0.id}-X-{op1.id}",
        source = op0.id,
        target = op1.id,
        type = "execution-order",
        nbox_string = None
    )
  )

  return {
    "flowchart": {
      "edges": [x.get_dict() for x in edges],
      "nodes": [x.get_dict() for x in nodes],
    },
    "symbols": symbols_to_nodes
  }



r"""This submodule concerns itself with conversion of different framworks to other frameworks.
It achieves this by providing a fix set of functions for each framework. There are a couple of
caveats that the developer must know about.

1. We use joblib to serialize the model, see `reason <https://stackoverflow.com/questions/12615525/what-are-the-different-use-cases-of-joblib-versus-pickle>`_ \
so when you will try to unpickle the model ``pickle`` will not work correctly and will throw the error
``_pickle.UnpicklingError: invalid load key, '\x00'``. So ensure that you use ``joblib``.

2. Serializing torch models directly is a bit tricky and weird, you can read more about it
`here <https://github.com/pytorch/pytorch/blob/master/torch/csrc/jit/docs/serialization.md>`_,
so technically pytorch torch.save() automatically pickles the object along with the required
datapoint (model hierarchy, constants, data, etc.)

Lazy Loading
------------

Lazy loading is a mechanism that allows you to load the model only when you need it, this is easier said than
done because you need to add many checks and balances at varaious locations in the code. The way it works here
is that we check from a list of modules which are required to import which part of the model.


Documentation
-------------
"""

# this function is for getting the meta data and is framework agnostic, so adding this in the
# __init__ of framework submodule
def get_meta(input_names, input_shapes, args, output_names, output_shapes, outputs):
    """Generic method to convert the inputs to get ``nbox_meta['metadata']`` dictionary"""
    # get the meta object
    def __get_struct(names_, shapes_, tensors_):
        return {
            name: {
                "dtype": str(tensor.dtype),
                "tensorShape": {"dim": [{"name": "", "size": x} for x in shapes], "unknownRank": False},
                "name": name,
            }
            for name, shapes, tensor in zip(names_, shapes_, tensors_)
        }

    meta = {"inputs": __get_struct(input_names, input_shapes, args), "outputs": __get_struct(output_names, output_shapes, outputs)}

    return meta


__all__ = ["get_meta"]

from types import SimpleNamespace
from ..utils import _isthere, folder, join

def update_all_lazy_loading(*modules, fname):
    """Lazy load modules and update the ``__all__`` list"""
    import sys
    global __all__
    if _isthere(*modules):
        sys.path.append(join(folder(__file__), f"{fname}.py"))
        _all_name = f"{fname}_all"
        eval(f"from .{fname} import *")
        eval(f"from .{fname} import __all__ as {_all_name}")
        maps = {f"{x}": globals()[x] for x in _all_name}
        maps.update({"IMPORTS": [*modules]})
        _name_of_module = fname.strip("_")
        eval(f"{_name_of_module} = SimpleNamespace(**maps)")
        __all__ += [f"{_name_of_module}"]

# update_all_lazy_loading("torch", fname = "__pytorch")  
# update_all_lazy_loading("sklearn", "sk2onnx", fname = "__sklearn")


_pt_modules = ["torch"]
if _isthere(*_pt_modules):
    from .__pytorch import *
    from .__pytorch import __all__ as _pt_all
    maps = {f"{x}": globals()[x] for x in _pt_all}
    maps.update({"IMPORTS": _pt_modules})
    pytorch = SimpleNamespace(**maps)
    __all__ += ["pytorch"]

_sk_modules = ["sklearn", "sk2onnx"]
if _isthere(*_sk_modules):
    from .__sklearn import *
    from .__sklearn import __all__ as _sk_all
    maps = {f"{x}": globals()[x] for x in _sk_all}
    maps.update({"IMPORTS": _sk_modules})
    sklearn = SimpleNamespace(**maps)
    __all__ += ["sklearn"]

_onnx_modules = ["onnx", "onnxruntime"]
if _isthere(*_onnx_modules):
    maps = {"IMPORTS": _onnx_modules}
    onnx = SimpleNamespace(**maps)
    __all__ += ["onnx"]



# this file has methods for netorking related things

import json
import os
import requests
from pprint import pprint as pp
from time import sleep

from . import utils

import logging
logger = logging.getLogger()

class NBXAPIError(Exception):
    pass


def deploy_model(
    export_model_path,
    nbox_meta,
    wait_for_deployment=False,
):
    """One-Click-Deploy method v3 that takes in a .nbox file and deploys it to the nbox server.
    Avoid using this function manually and use ``model.deploy()`` or nboxCLI instead.

    Args:
        export_model_path (str): path to the file to upload
        deployment_type (str, optional): type of deployment strategy
        nbox_meta (dict, optional): metadata for the nbox.Model() object being deployed
        wait_for_deployment (bool, optional): if true, acts like a blocking call (sync vs async)
        convert_args (str, optional): if deployment type == "ovms2" can pass extra arguments to MO
        deployment_id (str, optional): ``deployment_id`` to put this model under, if you do not pass this
            it will automatically create a new deployment check `platform <https://nimblebox.ai/oneclick>`_
            for more info or check the logs.
        deployment_name (str, optional): if ``deployment_id`` is not given and you want to create a new
            deployment group (ie. webserver will create a new ``deployment_id``) you can tell what name you
            want, be default it will create a random name.

    Returns:
        endpoint (str, None): if ``wait_for_deployment == True``, returns the URL endpoint of the deployed
            model
        access_key(str, None): if ``wait_for_deployment == True``, returns the data access key of
            the deployed model
    """
    from nbox.auth import secret  # it can refresh so add it in the method

    # pp(nbox_meta)

    access_token = secret.get("access_token")
    URL = secret.get("nbx_url")
    file_size = os.stat(export_model_path).st_size // (1024 ** 2)  # in MBs

    # intialise the console logger
    logger.info("-" * 30 + " NBX Deploy " + "-" * 30)
    logger.info(f"Deploying on URL: {URL}")
    deployment_type = nbox_meta["spec"]["deployment_type"]
    deployment_id = nbox_meta["spec"]["deployment_id"]
    deployment_name = nbox_meta["spec"]["deployment_name"]
    model_name = nbox_meta["spec"]["model_name"]
    
    logger.info(f"Deployment Type: '{deployment_type}', Deployment ID: '{deployment_id}'")

    if not deployment_id and not deployment_name:
        logger.info("Deployment ID not passed will create a new deployment with name >>")
        deployment_name = utils.get_random_name().replace("-", "_")

    logger.info(
        f"Deployment Name: '{deployment_name}', Model Name: '{model_name}', Model Path: '{export_model_path}', file_size: {file_size} MBs"
    )
    logger.info("Getting bucket URL")

    # get bucket URL
    r = requests.get(
        url=f"{URL}/api/model/get_upload_url",
        params={
            "file_size": file_size,  # because in MB
            "file_type": "nbox",
            "model_name": model_name,
            "convert_args": nbox_meta["spec"]["convert_args"],
            "nbox_meta": json.dumps(nbox_meta),  # annoying, but otherwise only the first key would be sent
            "deployment_type": deployment_type,  # "nbox" or "ovms2"
            "deployment_id": deployment_id,
            "deployment_name": deployment_name,
        },
        headers={"Authorization": f"Bearer {access_token}"},
    )
    try:
        r.raise_for_status()
    except:
        raise ValueError(f"Could not fetch upload URL: {r.content.decode('utf-8')}")
    out = r.json()
    model_id = out["fields"]["x-amz-meta-model_id"]
    deployment_id = out["fields"]["x-amz-meta-deployment_id"]
    logger.info(f"model_id: {model_id}")
    logger.info(f"deployment_id: {deployment_id}")

    # upload the file to a S3 -> don't raise for status here
    logger.info("Uploading model to S3 ...")
    r = requests.post(url=out["url"], data=out["fields"], files={"file": (out["fields"]["key"], open(export_model_path, "rb"))})

    # checking if file is successfully uploaded on S3 and tell webserver
    # whether upload is completed or not because client tells
    logger.info("Verifying upload ...")
    requests.post(
        url=f"{URL}/api/model/update_model_status",
        json={"upload": True if r.status_code == 204 else False, "model_id": model_id, "deployment_id": deployment_id},
        headers={"Authorization": f"Bearer {access_token}"},
    )

    # polling
    endpoint = None
    _stat_done = []  # status calls performed
    total_retries = 0  # number of hits it took
    access_key = None  # this key is used for calling the model
    logger.info(f"Check your deployment at {URL}/oneclick")
    if not wait_for_deployment:
        logger.info("NBX Deploy")
        return endpoint, access_key

    logger.info("Start Polling ...")
    while True:
        total_retries += 1

        # don't keep polling for very long, kill after sometime
        if total_retries > 50 and not wait_for_deployment:
            logger.info(f"Stopping polling, please check status at: {URL}/oneclick")
            break

        sleep(5)

        # get the status update
        logger.info(f"Getting updates ...")
        r = requests.get(
            url=f"{URL}/api/model/get_model_history",
            params={"model_id": model_id, "deployment_id": deployment_id},
            headers={"Authorization": f"Bearer {access_token}"},
        )
        try:
            r.raise_for_status()
            updates = r.json()
        except:
            pp(r.content)
            raise NBXAPIError("This should not happen, please raise an issue at https://github.com/NimbleBoxAI/nbox/issues with above log!")

        # go over all the status updates and check if the deployment is done
        for st in updates["model_history"]:
            curr_st = st["status"]
            if curr_st in _stat_done:
                continue

            # # only when this is a new status
            # col = {"failed": console.T.fail, "in-progress": console.T.inp, "success": console.T.st, "ready": console.T.st}[
            #     curr_st.split(".")[-1]
            # ]
            logger.info(f"Status: {curr_st}")
            _stat_done.append(curr_st)

        if curr_st == "deployment.success":
            # if we do not have api key then query web server for it
            if access_key is None:
                endpoint = updates["model_data"]["api_url"]

                if endpoint is None:
                    if wait_for_deployment:
                        continue
                    logger.info("Deployment in progress ...")
                    logger.info(f"Endpoint to be setup, please check status at: {URL}/oneclick")
                    break

                logger.info(f"Deployment successful at URL: {endpoint}")

                r = requests.get(
                    url=f"{URL}/api/model/get_deployment_access_key",
                    headers={"Authorization": f"Bearer {access_token}"},
                    params={"deployment_id": deployment_id},
                )
                try:
                    r.raise_for_status()
                    access_key = r.json()["access_key"]
                    logger.info(f"nbx-key: {access_key}")
                except:
                    pp(r.content.decode("utf-8"))
                    raise ValueError(f"Failed to get access_key, please check status at: {URL}/oneclick")

            # keep hitting /metadata and see if model is ready or not
            r = requests.get(url=f"{endpoint}/metadata", headers={"NBX-KEY": access_key, "Authorization": f"Bearer {access_token}"})
            if r.status_code == 200:
                logger.info(f"Model is ready")
                break

        # actual break condition happens here: bug in webserver where it does not return ready
        # curr_st == "ready"
        if access_key != None or "failed" in curr_st:
            break

    logger.info("Process Complete")
    logger.info("NBX Deploy")
    return endpoint, access_key


def deploy_job(
    zip_path: str,
    schedule_meta: dict,
):
    from nbox.auth import secret  # it can refresh so add it in the method

    access_token = secret.get("access_token")
    URL = secret.get("nbx_url")
    file_size = os.stat(zip_path).st_size // (1024 ** 2)  # in MBs

    # intialise the console logger
    logger.info("-" * 30 + " NBX Deploy " + "-" * 30)
    logger.info(f"Deploying on URL: {URL}")

    # POST not GET !change vs. model
    r = requests.post(
        url=f"{URL}/api/jobs/get_upload_url",
        headers={"Authorization": f"Bearer {access_token}"},
        json = {
            "file_size": file_size,  # because in MB
            "file_type": "zip",
            "job_name": schedule_meta["job_name"],
            "schedule_meta": schedule_meta,
        }
    )
    try:
        r.raise_for_status()
    except:
        raise ValueError(f"Could not fetch upload URL: {r.content.decode('utf-8')}")

    out = r.json()
    job_id = out["fields"]["x-amz-meta-job_id"]
    jobs_deployment_id = out["fields"]["x-amz-meta-jobs_deployment_id"]
    logger.info(f"job_id: {job_id}")
    logger.info(f"jobs_deployment_id: {jobs_deployment_id}")

    # upload the file to a S3 -> don't raise for status here
    logger.info("Uploading model to S3 ...")
    r = requests.post(url=out["url"], data=out["fields"], files={"file": (out["fields"]["key"], open(zip_path, "rb"))})

    # checking if file is successfully uploaded on S3 and tell webserver whether upload is completed or not because client tells
    logger.info("Verifying upload ...")
    r = requests.post(
        url=f"{URL}/api/jobs/update_model_status",
        json={"upload": True if r.status_code == 204 else False, "job_id": job_id, "jobs_deployment_id": jobs_deployment_id},
        headers={"Authorization": f"Bearer {access_token}"},
    )
    try:
        r.raise_for_status()
    except:
        raise ValueError(f"Could not update model status: {r.content.decode('utf-8')}")


